# -*- coding: utf-8 -*-
'''
Tests for the fileserver runner
'''
# Import Python libs
from __future__ import absolute_import
import contextlib

# Import Salt Testing libs
from salttesting.helpers import ensure_in_syspath
ensure_in_syspath('../../')

# Import salt libs
import integration


class FileserverTest(integration.ShellCase):
    '''
    Test the fileserver runner
    '''
    def test_dir_list(self):
        '''
        fileserver.dir_list
        '''
        ret = self.run_run_plus(fun='fileserver.dir_list')
        self.assertIsInstance(ret['return'], list)
        self.assertTrue('_modules' in ret['return'])

        # Backend submitted as a string
        ret = self.run_run_plus(fun='fileserver.dir_list', backend='roots')
        self.assertIsInstance(ret['return'], list)
        self.assertTrue('_modules' in ret['return'])

        # Backend submitted as a list
        ret = self.run_run_plus(fun='fileserver.dir_list', backend=['roots'])
        self.assertIsInstance(ret['return'], list)
        self.assertTrue('_modules' in ret['return'])

    def test_empty_dir_list(self):
        '''
        fileserver.empty_dir_list
        '''
        ret = self.run_run_plus(fun='fileserver.empty_dir_list')
        self.assertIsInstance(ret['return'], list)
        self.assertEqual(ret['return'], [])

        # Backend submitted as a string
        ret = self.run_run_plus(
            fun='fileserver.empty_dir_list',
            backend='roots')
        self.assertIsInstance(ret['return'], list)
        self.assertEqual(ret['return'], [])

        # Backend submitted as a list
        ret = self.run_run_plus(
            fun='fileserver.empty_dir_list',
            backend=['roots'])
        self.assertIsInstance(ret['return'], list)
        self.assertEqual(ret['return'], [])

    def test_envs(self):
        '''
        fileserver.envs
        '''
        ret = self.run_run_plus(fun='fileserver.envs')
        self.assertIsInstance(ret['return'], list)

        # Backend submitted as a string
        ret = self.run_run_plus(fun='fileserver.envs', backend='roots')
        self.assertIsInstance(ret['return'], list)

        # Backend submitted as a list
        ret = self.run_run_plus(fun='fileserver.envs', backend=['roots'])
        self.assertIsInstance(ret['return'], list)

    def test_clear_file_list_cache(self):
        '''
        fileserver.clear_file_list_cache

        If this test fails, then something may have changed in the test suite
        and we may have more than just "roots" configured in the fileserver
        backends. This assert will need to be updated accordingly.
        '''
        saltenvs = sorted(self.run_run_plus(fun='fileserver.envs')['return'])

        @contextlib.contextmanager
        def gen_cache():
            '''
            Create file_list cache so we have something to clear
            '''
            for saltenv in saltenvs:
                self.run_run_plus(fun='fileserver.file_list', saltenv=saltenv)
            yield

        # Test with no arguments
        with gen_cache():
            ret = self.run_run_plus(fun='fileserver.clear_file_list_cache')
            ret['return']['roots'].sort()
            self.assertEqual(ret['return'], {'roots': saltenvs})

        # Test with backend passed as string
        with gen_cache():
            ret = self.run_run_plus(fun='fileserver.clear_file_list_cache',
                                    backend='roots')
            ret['return']['roots'].sort()
            self.assertEqual(ret['return'], {'roots': saltenvs})

        # Test with backend passed as list
        with gen_cache():
            ret = self.run_run_plus(fun='fileserver.clear_file_list_cache',
                                    backend=['roots'])
            ret['return']['roots'].sort()
            self.assertEqual(ret['return'], {'roots': saltenvs})

        # Test with backend passed as string, but with a nonsense backend
        with gen_cache():
            ret = self.run_run_plus(fun='fileserver.clear_file_list_cache',
                                    backend='notarealbackend')
            self.assertEqual(ret['return'], {})

        # Test with saltenv passed as string
        with gen_cache():
            ret = self.run_run_plus(fun='fileserver.clear_file_list_cache',
                                    saltenv='base')
            self.assertEqual(ret['return'], {'roots': ['base']})

        # Test with saltenv passed as list
        with gen_cache():
            ret = self.run_run_plus(fun='fileserver.clear_file_list_cache',
                                    saltenv=['base'])
            self.assertEqual(ret['return'], {'roots': ['base']})

        # Test with saltenv passed as string, but with a nonsense saltenv
        with gen_cache():
            ret = self.run_run_plus(fun='fileserver.clear_file_list_cache',
                                    saltenv='notarealsaltenv')
            self.assertEqual(ret['return'], {})

        # Test with both backend and saltenv passed
        with gen_cache():
            ret = self.run_run_plus(fun='fileserver.clear_file_list_cache',
                                    backend='roots',
                                    saltenv='base')
            self.assertEqual(ret['return'], {'roots': ['base']})

    def test_file_list(self):
        '''
        fileserver.file_list
        '''
        ret = self.run_run_plus(fun='fileserver.file_list')
        self.assertIsInstance(ret['return'], list)
        self.assertTrue('grail/scene33' in ret['return'])

        # Backend submitted as a string
        ret = self.run_run_plus(fun='fileserver.file_list', backend='roots')
        self.assertIsInstance(ret['return'], list)
        self.assertTrue('grail/scene33' in ret['return'])

        # Backend submitted as a list
        ret = self.run_run_plus(fun='fileserver.file_list', backend=['roots'])
        self.assertIsInstance(ret['return'], list)
        self.assertTrue('grail/scene33' in ret['return'])

    def test_symlink_list(self):
        '''
        fileserver.symlink_list
        '''
        ret = self.run_run_plus(fun='fileserver.symlink_list')
        self.assertIsInstance(ret['return'], dict)
        self.assertTrue('dest_sym' in ret['return'])

        # Backend submitted as a string
        ret = self.run_run_plus(fun='fileserver.symlink_list', backend='roots')
        self.assertIsInstance(ret['return'], dict)
        self.assertTrue('dest_sym' in ret['return'])

        # Backend submitted as a list
        ret = self.run_run_plus(fun='fileserver.symlink_list', backend=['roots'])
        self.assertIsInstance(ret['return'], dict)
        self.assertTrue('dest_sym' in ret['return'])

    def test_update(self):
        '''
        fileserver.update
        '''
        ret = self.run_run_plus(fun='fileserver.update')
        self.assertTrue(ret['return'])

        # Backend submitted as a string
        ret = self.run_run_plus(fun='fileserver.update', backend='roots')
        self.assertTrue(ret['return'])

        # Backend submitted as a list
        ret = self.run_run_plus(fun='fileserver.update', backend=['roots'])
        self.assertTrue(ret['return'])

if __name__ == '__main__':
    from integration import run_tests
    run_tests(FileserverTest)



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import
from distutils.version import LooseVersion  # pylint: disable=import-error,no-name-in-module
import random
import string

# Import Salt Testing libs
from salttesting.unit import skipIf, TestCase
from salttesting.mock import (
    MagicMock,
    NO_MOCK,
    NO_MOCK_REASON,
    patch
)
from salttesting.helpers import ensure_in_syspath

ensure_in_syspath('../../')

# Import Salt libs
import salt.config
import salt.loader
from salt.modules import boto_iot
from salt.ext.six.moves import range  # pylint: disable=import-error,redefined-builtin

# Import 3rd-party libs
import logging

# pylint: disable=import-error,no-name-in-module,unused-import
try:
    import boto
    import boto3
    from botocore.exceptions import ClientError
    from botocore import __version__ as found_botocore_version
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

# pylint: enable=import-error,no-name-in-module,unused-import

# the boto_iot module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto3_version = '1.2.1'
required_botocore_version = '1.4.41'

log = logging.getLogger(__name__)

opts = salt.config.DEFAULT_MINION_OPTS
context = {}
utils = salt.loader.utils(opts, whitelist=['boto3'], context=context)

boto_iot.__utils__ = utils
boto_iot.__init__(opts)
boto_iot.__salt__ = {}


def _has_required_boto():
    '''
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    '''
    if not HAS_BOTO:
        return False
    elif LooseVersion(boto3.__version__) < LooseVersion(required_boto3_version):
        return False
    elif LooseVersion(found_botocore_version) < LooseVersion(required_botocore_version):
        return False
    else:
        return True


if _has_required_boto():
    region = 'us-east-1'
    access_key = 'GKTADJGHEIQSXMKKRBJ08H'
    secret_key = 'askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs'
    conn_parameters = {'region': region, 'key': access_key, 'keyid': secret_key, 'profile': {}}
    error_message = 'An error occurred (101) when calling the {0} operation: Test-defined error'
    not_found_error = ClientError({
        'Error': {
            'Code': 'ResourceNotFoundException',
            'Message': "Test-defined error"
        }
    }, 'msg')
    topic_rule_not_found_error = ClientError({
        'Error': {
            'Code': 'UnauthorizedException',
            'Message': "Test-defined error"
        }
    }, 'msg')
    error_content = {
      'Error': {
        'Code': 101,
        'Message': "Test-defined error"
      }
    }
    policy_ret = dict(policyName='testpolicy',
                      policyDocument='{"Version": "2012-10-17", "Statement": [{"Action": ["iot:Publish"], "Resource": ["*"], "Effect": "Allow"}]}',
                      policyArn='arn:aws:iot:us-east-1:123456:policy/my_policy',
                      policyVersionId=1,
                      defaultVersionId=1)
    topic_rule_ret = dict(ruleName='testrule',
                      sql="SELECT * FROM 'iot/test'",
                      description='topic rule description',
                      createdAt='1970-01-01',
                      actions=[{'lambda': {'functionArn': 'arn:aws:::function'}}],
                      ruleDisabled=True)

    thing_type_name = 'test_thing_type'
    thing_type_desc = 'test_thing_type_desc'
    thing_type_attr_1 = 'test_thing_type_search_attr_1'
    thing_type_ret = dict(
        thingTypeName=thing_type_name,
        thingTypeProperties=dict(
            thingTypeDescription=thing_type_desc,
            searchableAttributes=[thing_type_attr_1],
        ),
        thingTypeMetadata=dict(
            deprecated=False,
            creationDate='test_thing_type_create_date'
        )
    )
    thing_type_arn = 'test_thing_type_arn'
    create_thing_type_ret = dict(
        thingTypeName=thing_type_name,
        thingTypeArn=thing_type_arn
    )


@skipIf(HAS_BOTO is False, 'The boto module must be installed.')
@skipIf(_has_required_boto() is False, 'The boto3 module must be greater than'
                                       ' or equal to version {0}'
        .format(required_boto3_version))
@skipIf(NO_MOCK, NO_MOCK_REASON)
class BotoIoTTestCaseBase(TestCase):
    conn = None

    # Set up MagicMock to replace the boto3 session
    def setUp(self):
        boto_iot.__context__ = {}
        context.clear()
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters['key'] = ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(50))

        self.patcher = patch('boto3.session.Session')
        self.addCleanup(self.patcher.stop)
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn = MagicMock()
        session_instance.client.return_value = self.conn


class BotoIoTTestCaseMixin(object):
    pass


@skipIf(HAS_BOTO is False, 'The boto module must be installed.')
@skipIf(_has_required_boto() is False, 'The boto3 module must be greater than'
                                       ' or equal to version {0}.  The botocore'
                                       ' module must be greater than or equal to'
                                       ' version {1}.'
        .format(required_boto3_version, required_botocore_version))
@skipIf(NO_MOCK, NO_MOCK_REASON)
class BotoIoTThingTypeTestCase(BotoIoTTestCaseBase, BotoIoTTestCaseMixin):
    '''
    TestCase for salt.modules.boto_iot module
    '''

    def test_that_when_checking_if_a_thing_type_exists_and_a_thing_type_exists_the_thing_type_exists_method_returns_true(self):
        '''
        Tests checking iot thing type existence when the iot thing type already exists
        '''
        self.conn.describe_thing_type.return_value = thing_type_ret
        result = boto_iot.thing_type_exists(thingTypeName=thing_type_name, **conn_parameters)

        self.assertTrue(result['exists'])

    def test_that_when_checking_if_a_thing_type_exists_and_a_thing_type_does_not_exist_the_thing_type_exists_method_returns_false(self):
        '''
        Tests checking iot thing type existence when the iot thing type does not exist
        '''
        self.conn.describe_thing_type.side_effect = not_found_error
        result = boto_iot.thing_type_exists(thingTypeName='non existent thing type', **conn_parameters)

        self.assertFalse(result['exists'])

    def test_that_when_checking_if_a_thing_type_exists_and_boto3_returns_an_error_the_thing_type_exists_method_returns_error(self):
        '''
        Tests checking iot thing type existence when boto returns an error
        '''
        self.conn.describe_thing_type.side_effect = ClientError(error_content, 'describe_thing_type')
        result = boto_iot.thing_type_exists(thingTypeName='mythingtype', **conn_parameters)

        self.assertEqual(result.get('error', {}).get('message'), error_message.format('describe_thing_type'))

    def test_that_when_describing_thing_type_and_thing_type_exists_the_describe_thing_type_method_returns_thing_type(self):
        '''
        Tests describe thing type for an existing thing type
        '''
        self.conn.describe_thing_type.return_value = thing_type_ret
        result = boto_iot.describe_thing_type(thingTypeName=thing_type_name, **conn_parameters)

        self.assertEqual(result.get('thing_type'), thing_type_ret)

    def test_that_when_describing_thing_type_and_thing_type_does_not_exists_the_describe_thing_type_method_returns_none(self):
        '''
        Tests describe thing type for an non existent thing type
        '''
        self.conn.describe_thing_type.side_effect = not_found_error
        result = boto_iot.describe_thing_type(thingTypeName='non existent thing type', **conn_parameters)

        self.assertEqual(result.get('thing_type'), None)

    def test_that_when_describing_thing_type_and_boto3_returns_error_an_error_the_describe_thing_type_method_returns_error(self):
        self.conn.describe_thing_type.side_effect = ClientError(error_content, 'describe_thing_type')
        result = boto_iot.describe_thing_type(thingTypeName='mythingtype', **conn_parameters)

        self.assertEqual(result.get('error', {}).get('message'), error_message.format('describe_thing_type'))

    def test_that_when_creating_a_thing_type_succeeds_the_create_thing_type_method_returns_true(self):
        '''
        tests True when thing type created
        '''
        self.conn.create_thing_type.return_value = create_thing_type_ret
        result = boto_iot.create_thing_type(thingTypeName=thing_type_name,
                                            thingTypeDescription=thing_type_desc,
                                            searchableAttributesList=[thing_type_attr_1],
                                            **conn_parameters)
        self.assertTrue(result['created'])
        self.assertTrue(result['thingTypeArn'], thing_type_arn)

    def test_that_when_creating_a_thing_type_fails_the_create_thing_type_method_returns_error(self):
        '''
        tests False when thing type not created
        '''
        self.conn.create_thing_type.side_effect = ClientError(error_content, 'create_thing_type')
        result = boto_iot.create_thing_type(thingTypeName=thing_type_name,
                                            thingTypeDescription=thing_type_desc,
                                            searchableAttributesList=[thing_type_attr_1],
                                            **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('create_thing_type'))

    def test_that_when_deprecating_a_thing_type_succeeds_the_deprecate_thing_type_method_returns_true(self):
        '''
        tests True when deprecate thing type succeeds
        '''
        self.conn.deprecate_thing_type.return_value = {}
        result = boto_iot.deprecate_thing_type(thingTypeName=thing_type_name,
                                               undoDeprecate=False,
                                               **conn_parameters)

        self.assertTrue(result.get('deprecated'))
        self.assertEqual(result.get('error'), None)

    def test_that_when_deprecating_a_thing_type_fails_the_deprecate_thing_type_method_returns_error(self):
        '''
        tests False when thing type fails to deprecate
        '''
        self.conn.deprecate_thing_type.side_effect = ClientError(error_content, 'deprecate_thing_type')
        result = boto_iot.deprecate_thing_type(thingTypeName=thing_type_name,
                                               undoDeprecate=False,
                                               **conn_parameters)

        self.assertFalse(result.get('deprecated'))
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('deprecate_thing_type'))

    def test_that_when_deleting_a_thing_type_succeeds_the_delete_thing_type_method_returns_true(self):
        '''
        tests True when delete thing type succeeds
        '''
        self.conn.delete_thing_type.return_value = {}
        result = boto_iot.delete_thing_type(thingTypeName=thing_type_name, **conn_parameters)

        self.assertTrue(result.get('deleted'))
        self.assertEqual(result.get('error'), None)

    def test_that_when_deleting_a_thing_type_fails_the_delete_thing_type_method_returns_error(self):
        '''
        tests False when delete thing type fails
        '''
        self.conn.delete_thing_type.side_effect = ClientError(error_content, 'delete_thing_type')
        result = boto_iot.delete_thing_type(thingTypeName=thing_type_name, **conn_parameters)
        self.assertFalse(result.get('deleted'))
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('delete_thing_type'))


@skipIf(HAS_BOTO is False, 'The boto module must be installed.')
@skipIf(_has_required_boto() is False, 'The boto3 module must be greater than'
                                       ' or equal to version {0}.  The botocore'
                                       ' module must be greater than or equal to'
                                       ' version {1}.'
        .format(required_boto3_version, required_botocore_version))
@skipIf(NO_MOCK, NO_MOCK_REASON)
class BotoIoTPolicyTestCase(BotoIoTTestCaseBase, BotoIoTTestCaseMixin):
    '''
    TestCase for salt.modules.boto_iot module
    '''

    def test_that_when_checking_if_a_policy_exists_and_a_policy_exists_the_policy_exists_method_returns_true(self):
        '''
        Tests checking iot policy existence when the iot policy already exists
        '''
        self.conn.get_policy.return_value = {'policy': policy_ret}
        result = boto_iot.policy_exists(policyName=policy_ret['policyName'], **conn_parameters)

        self.assertTrue(result['exists'])

    def test_that_when_checking_if_a_policy_exists_and_a_policy_does_not_exist_the_policy_exists_method_returns_false(self):
        '''
        Tests checking iot policy existence when the iot policy does not exist
        '''
        self.conn.get_policy.side_effect = not_found_error
        result = boto_iot.policy_exists(policyName='mypolicy', **conn_parameters)

        self.assertFalse(result['exists'])

    def test_that_when_checking_if_a_policy_exists_and_boto3_returns_an_error_the_policy_exists_method_returns_error(self):
        '''
        Tests checking iot policy existence when boto returns an error
        '''
        self.conn.get_policy.side_effect = ClientError(error_content, 'get_policy')
        result = boto_iot.policy_exists(policyName='mypolicy', **conn_parameters)

        self.assertEqual(result.get('error', {}).get('message'), error_message.format('get_policy'))

    def test_that_when_creating_a_policy_succeeds_the_create_policy_method_returns_true(self):
        '''
        tests True policy created.
        '''
        self.conn.create_policy.return_value = policy_ret
        result = boto_iot.create_policy(policyName=policy_ret['policyName'],
                                        policyDocument=policy_ret['policyDocument'],
                                        **conn_parameters)

        self.assertTrue(result['created'])

    def test_that_when_creating_a_policy_fails_the_create_policy_method_returns_error(self):
        '''
        tests False policy not created.
        '''
        self.conn.create_policy.side_effect = ClientError(error_content, 'create_policy')
        result = boto_iot.create_policy(policyName=policy_ret['policyName'],
                                        policyDocument=policy_ret['policyDocument'],
                                        **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('create_policy'))

    def test_that_when_deleting_a_policy_succeeds_the_delete_policy_method_returns_true(self):
        '''
        tests True policy deleted.
        '''
        result = boto_iot.delete_policy(policyName='testpolicy',
                                        **conn_parameters)

        self.assertTrue(result['deleted'])

    def test_that_when_deleting_a_policy_fails_the_delete_policy_method_returns_false(self):
        '''
        tests False policy not deleted.
        '''
        self.conn.delete_policy.side_effect = ClientError(error_content, 'delete_policy')
        result = boto_iot.delete_policy(policyName='testpolicy',
                                        **conn_parameters)
        self.assertFalse(result['deleted'])

    def test_that_when_describing_policy_it_returns_the_dict_of_properties_returns_true(self):
        '''
        Tests describing parameters if policy exists
        '''
        self.conn.get_policy.return_value = {'policy': policy_ret}

        result = boto_iot.describe_policy(policyName=policy_ret['policyName'], **conn_parameters)

        self.assertTrue(result['policy'])

    def test_that_when_describing_policy_it_returns_the_dict_of_properties_returns_false(self):
        '''
        Tests describing parameters if policy does not exist
        '''
        self.conn.get_policy.side_effect = not_found_error
        result = boto_iot.describe_policy(policyName='testpolicy', **conn_parameters)

        self.assertFalse(result['policy'])

    def test_that_when_describing_policy_on_client_error_it_returns_error(self):
        '''
        Tests describing parameters failure
        '''
        self.conn.get_policy.side_effect = ClientError(error_content, 'get_policy')
        result = boto_iot.describe_policy(policyName='testpolicy', **conn_parameters)
        self.assertTrue('error' in result)

    def test_that_when_checking_if_a_policy_version_exists_and_a_policy_version_exists_the_policy_version_exists_method_returns_true(self):
        '''
        Tests checking iot policy existence when the iot policy version already exists
        '''
        self.conn.get_policy.return_value = {'policy': policy_ret}
        result = boto_iot.policy_version_exists(policyName=policy_ret['policyName'],
                                                policyVersionId=1,
                                                **conn_parameters)

        self.assertTrue(result['exists'])

    def test_that_when_checking_if_a_policy_version_exists_and_a_policy_version_does_not_exist_the_policy_version_exists_method_returns_false(self):
        '''
        Tests checking iot policy_version existence when the iot policy_version does not exist
        '''
        self.conn.get_policy_version.side_effect = not_found_error
        result = boto_iot.policy_version_exists(policyName=policy_ret['policyName'],
                                                policyVersionId=1,
                                                **conn_parameters)

        self.assertFalse(result['exists'])

    def test_that_when_checking_if_a_policy_version_exists_and_boto3_returns_an_error_the_policy_version_exists_method_returns_error(self):
        '''
        Tests checking iot policy_version existence when boto returns an error
        '''
        self.conn.get_policy_version.side_effect = ClientError(error_content, 'get_policy_version')
        result = boto_iot.policy_version_exists(policyName=policy_ret['policyName'],
                                                policyVersionId=1,
                                                **conn_parameters)

        self.assertEqual(result.get('error', {}).get('message'), error_message.format('get_policy_version'))

    def test_that_when_creating_a_policy_version_succeeds_the_create_policy_version_method_returns_true(self):
        '''
        tests True policy_version created.
        '''
        self.conn.create_policy_version.return_value = policy_ret
        result = boto_iot.create_policy_version(policyName=policy_ret['policyName'],
                                        policyDocument=policy_ret['policyDocument'],
                                        **conn_parameters)

        self.assertTrue(result['created'])

    def test_that_when_creating_a_policy_version_fails_the_create_policy_version_method_returns_error(self):
        '''
        tests False policy_version not created.
        '''
        self.conn.create_policy_version.side_effect = ClientError(error_content, 'create_policy_version')
        result = boto_iot.create_policy_version(policyName=policy_ret['policyName'],
                                        policyDocument=policy_ret['policyDocument'],
                                        **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('create_policy_version'))

    def test_that_when_deleting_a_policy_version_succeeds_the_delete_policy_version_method_returns_true(self):
        '''
        tests True policy_version deleted.
        '''
        result = boto_iot.delete_policy_version(policyName='testpolicy',
                                        policyVersionId=1,
                                        **conn_parameters)

        self.assertTrue(result['deleted'])

    def test_that_when_deleting_a_policy_version_fails_the_delete_policy_version_method_returns_false(self):
        '''
        tests False policy_version not deleted.
        '''
        self.conn.delete_policy_version.side_effect = ClientError(error_content, 'delete_policy_version')
        result = boto_iot.delete_policy_version(policyName='testpolicy',
                                        policyVersionId=1,
                                        **conn_parameters)
        self.assertFalse(result['deleted'])

    def test_that_when_describing_policy_version_it_returns_the_dict_of_properties_returns_true(self):
        '''
        Tests describing parameters if policy_version exists
        '''
        self.conn.get_policy_version.return_value = {'policy': policy_ret}

        result = boto_iot.describe_policy_version(policyName=policy_ret['policyName'],
                                        policyVersionId=1,
                                        **conn_parameters)

        self.assertTrue(result['policy'])

    def test_that_when_describing_policy_version_it_returns_the_dict_of_properties_returns_false(self):
        '''
        Tests describing parameters if policy_version does not exist
        '''
        self.conn.get_policy_version.side_effect = not_found_error
        result = boto_iot.describe_policy_version(policyName=policy_ret['policyName'],
                                        policyVersionId=1,
                                        **conn_parameters)

        self.assertFalse(result['policy'])

    def test_that_when_describing_policy_version_on_client_error_it_returns_error(self):
        '''
        Tests describing parameters failure
        '''
        self.conn.get_policy_version.side_effect = ClientError(error_content, 'get_policy_version')
        result = boto_iot.describe_policy_version(policyName=policy_ret['policyName'],
                                        policyVersionId=1,
                                        **conn_parameters)
        self.assertTrue('error' in result)

    def test_that_when_listing_policies_succeeds_the_list_policies_method_returns_true(self):
        '''
        tests True policies listed.
        '''
        self.conn.list_policies.return_value = {'policies': [policy_ret]}
        result = boto_iot.list_policies(**conn_parameters)

        self.assertTrue(result['policies'])

    def test_that_when_listing_policy_fails_the_list_policy_method_returns_false(self):
        '''
        tests False no policy listed.
        '''
        self.conn.list_policies.return_value = {'policies': []}
        result = boto_iot.list_policies(**conn_parameters)
        self.assertFalse(result['policies'])

    def test_that_when_listing_policy_fails_the_list_policy_method_returns_error(self):
        '''
        tests False policy error.
        '''
        self.conn.list_policies.side_effect = ClientError(error_content, 'list_policies')
        result = boto_iot.list_policies(**conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('list_policies'))

    def test_that_when_listing_policy_versions_succeeds_the_list_policy_versions_method_returns_true(self):
        '''
        tests True policy versions listed.
        '''
        self.conn.list_policy_versions.return_value = {'policyVersions': [policy_ret]}
        result = boto_iot.list_policy_versions(policyName='testpolicy',
                                               **conn_parameters)

        self.assertTrue(result['policyVersions'])

    def test_that_when_listing_policy_versions_fails_the_list_policy_versions_method_returns_false(self):
        '''
        tests False no policy versions listed.
        '''
        self.conn.list_policy_versions.return_value = {'policyVersions': []}
        result = boto_iot.list_policy_versions(policyName='testpolicy',
                                               **conn_parameters)
        self.assertFalse(result['policyVersions'])

    def test_that_when_listing_policy_versions_fails_the_list_policy_versions_method_returns_error(self):
        '''
        tests False policy versions error.
        '''
        self.conn.list_policy_versions.side_effect = ClientError(error_content, 'list_policy_versions')
        result = boto_iot.list_policy_versions(policyName='testpolicy',
                                               **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('list_policy_versions'))

    def test_that_when_setting_default_policy_version_succeeds_the_set_default_policy_version_method_returns_true(self):
        '''
        tests True policy version set.
        '''
        result = boto_iot.set_default_policy_version(policyName='testpolicy',
                                               policyVersionId=1,
                                               **conn_parameters)

        self.assertTrue(result['changed'])

    def test_that_when_set_default_policy_version_fails_the_set_default_policy_version_method_returns_error(self):
        '''
        tests False policy version error.
        '''
        self.conn.set_default_policy_version.side_effect = \
                                ClientError(error_content, 'set_default_policy_version')
        result = boto_iot.set_default_policy_version(policyName='testpolicy',
                                               policyVersionId=1,
                                               **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'),
                         error_message.format('set_default_policy_version'))

    def test_that_when_list_principal_policies_succeeds_the_list_principal_policies_method_returns_true(self):
        '''
        tests True policies listed.
        '''
        self.conn.list_principal_policies.return_value = {'policies': [policy_ret]}
        result = boto_iot.list_principal_policies(principal='us-east-1:GUID-GUID-GUID',
                                               **conn_parameters)

        self.assertTrue(result['policies'])

    def test_that_when_list_principal_policies_fails_the_list_principal_policies_method_returns_error(self):
        '''
        tests False policy version error.
        '''
        self.conn.list_principal_policies.side_effect = \
                                ClientError(error_content, 'list_principal_policies')
        result = boto_iot.list_principal_policies(principal='us-east-1:GUID-GUID-GUID',
                                               **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'),
                         error_message.format('list_principal_policies'))

    def test_that_when_attach_principal_policy_succeeds_the_attach_principal_policy_method_returns_true(self):
        '''
        tests True policy attached.
        '''
        result = boto_iot.attach_principal_policy(policyName='testpolicy',
                                               principal='us-east-1:GUID-GUID-GUID',
                                               **conn_parameters)

        self.assertTrue(result['attached'])

    def test_that_when_attach_principal_policy_version_fails_the_attach_principal_policy_version_method_returns_error(self):
        '''
        tests False policy version error.
        '''
        self.conn.attach_principal_policy.side_effect = \
                                ClientError(error_content, 'attach_principal_policy')
        result = boto_iot.attach_principal_policy(policyName='testpolicy',
                                               principal='us-east-1:GUID-GUID-GUID',
                                               **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'),
                         error_message.format('attach_principal_policy'))

    def test_that_when_detach_principal_policy_succeeds_the_detach_principal_policy_method_returns_true(self):
        '''
        tests True policy detached.
        '''
        result = boto_iot.detach_principal_policy(policyName='testpolicy',
                                               principal='us-east-1:GUID-GUID-GUID',
                                               **conn_parameters)

        self.assertTrue(result['detached'])

    def test_that_when_detach_principal_policy_version_fails_the_detach_principal_policy_version_method_returns_error(self):
        '''
        tests False policy version error.
        '''
        self.conn.detach_principal_policy.side_effect = \
                                ClientError(error_content, 'detach_principal_policy')
        result = boto_iot.detach_principal_policy(policyName='testpolicy',
                                               principal='us-east-1:GUID-GUID-GUID',
                                               **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'),
                         error_message.format('detach_principal_policy'))


@skipIf(HAS_BOTO is False, 'The boto module must be installed.')
@skipIf(_has_required_boto() is False, 'The boto3 module must be greater than'
                                       ' or equal to version {0}.  The botocore'
                                       ' module must be greater than or equal to'
                                       ' version {1}.'
        .format(required_boto3_version, required_botocore_version))
@skipIf(NO_MOCK, NO_MOCK_REASON)
class BotoIoTTopicRuleTestCase(BotoIoTTestCaseBase, BotoIoTTestCaseMixin):
    '''
    TestCase for salt.modules.boto_iot module
    '''

    def test_that_when_checking_if_a_topic_rule_exists_and_a_topic_rule_exists_the_topic_rule_exists_method_returns_true(self):
        '''
        Tests checking iot topic_rule existence when the iot topic_rule already exists
        '''
        self.conn.get_topic_rule.return_value = {'rule': topic_rule_ret}
        result = boto_iot.topic_rule_exists(ruleName=topic_rule_ret['ruleName'],
                                            **conn_parameters)

        self.assertTrue(result['exists'])

    def test_that_when_checking_if_a_rule_exists_and_a_rule_does_not_exist_the_topic_rule_exists_method_returns_false(self):
        '''
        Tests checking iot rule existence when the iot rule does not exist
        '''
        self.conn.get_topic_rule.side_effect = topic_rule_not_found_error
        result = boto_iot.topic_rule_exists(ruleName='mypolicy', **conn_parameters)

        self.assertFalse(result['exists'])

    def test_that_when_checking_if_a_topic_rule_exists_and_boto3_returns_an_error_the_topic_rule_exists_method_returns_error(self):
        '''
        Tests checking iot topic_rule existence when boto returns an error
        '''
        self.conn.get_topic_rule.side_effect = ClientError(error_content, 'get_topic_rule')
        result = boto_iot.topic_rule_exists(ruleName='myrule', **conn_parameters)

        self.assertEqual(result.get('error', {}).get('message'),
                                      error_message.format('get_topic_rule'))

    def test_that_when_creating_a_topic_rule_succeeds_the_create_topic_rule_method_returns_true(self):
        '''
        tests True topic_rule created.
        '''
        self.conn.create_topic_rule.return_value = topic_rule_ret
        result = boto_iot.create_topic_rule(ruleName=topic_rule_ret['ruleName'],
                                        sql=topic_rule_ret['sql'],
                                        description=topic_rule_ret['description'],
                                        actions=topic_rule_ret['actions'],
                                        **conn_parameters)

        self.assertTrue(result['created'])

    def test_that_when_creating_a_topic_rule_fails_the_create_topic_rule_method_returns_error(self):
        '''
        tests False topic_rule not created.
        '''
        self.conn.create_topic_rule.side_effect = ClientError(error_content, 'create_topic_rule')
        result = boto_iot.create_topic_rule(ruleName=topic_rule_ret['ruleName'],
                                        sql=topic_rule_ret['sql'],
                                        description=topic_rule_ret['description'],
                                        actions=topic_rule_ret['actions'],
                                        **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('create_topic_rule'))

    def test_that_when_replacing_a_topic_rule_succeeds_the_replace_topic_rule_method_returns_true(self):
        '''
        tests True topic_rule replaced.
        '''
        self.conn.replace_topic_rule.return_value = topic_rule_ret
        result = boto_iot.replace_topic_rule(ruleName=topic_rule_ret['ruleName'],
                                        sql=topic_rule_ret['sql'],
                                        description=topic_rule_ret['description'],
                                        actions=topic_rule_ret['actions'],
                                        **conn_parameters)

        self.assertTrue(result['replaced'])

    def test_that_when_replacing_a_topic_rule_fails_the_replace_topic_rule_method_returns_error(self):
        '''
        tests False topic_rule not replaced.
        '''
        self.conn.replace_topic_rule.side_effect = ClientError(error_content, 'replace_topic_rule')
        result = boto_iot.replace_topic_rule(ruleName=topic_rule_ret['ruleName'],
                                        sql=topic_rule_ret['sql'],
                                        description=topic_rule_ret['description'],
                                        actions=topic_rule_ret['actions'],
                                        **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('replace_topic_rule'))

    def test_that_when_deleting_a_topic_rule_succeeds_the_delete_topic_rule_method_returns_true(self):
        '''
        tests True topic_rule deleted.
        '''
        result = boto_iot.delete_topic_rule(ruleName='testrule',
                                        **conn_parameters)

        self.assertTrue(result['deleted'])

    def test_that_when_deleting_a_topic_rule_fails_the_delete_topic_rule_method_returns_false(self):
        '''
        tests False topic_rule not deleted.
        '''
        self.conn.delete_topic_rule.side_effect = ClientError(error_content, 'delete_topic_rule')
        result = boto_iot.delete_topic_rule(ruleName='testrule',
                                        **conn_parameters)
        self.assertFalse(result['deleted'])

    def test_that_when_describing_topic_rule_it_returns_the_dict_of_properties_returns_true(self):
        '''
        Tests describing parameters if topic_rule exists
        '''
        self.conn.get_topic_rule.return_value = {'rule': topic_rule_ret}

        result = boto_iot.describe_topic_rule(ruleName=topic_rule_ret['ruleName'],
                                              **conn_parameters)

        self.assertTrue(result['rule'])

    def test_that_when_describing_topic_rule_on_client_error_it_returns_error(self):
        '''
        Tests describing parameters failure
        '''
        self.conn.get_topic_rule.side_effect = ClientError(error_content, 'get_topic_rule')
        result = boto_iot.describe_topic_rule(ruleName='testrule', **conn_parameters)
        self.assertTrue('error' in result)

    def test_that_when_listing_topic_rules_succeeds_the_list_topic_rules_method_returns_true(self):
        '''
        tests True topic_rules listed.
        '''
        self.conn.list_topic_rules.return_value = {'rules': [topic_rule_ret]}
        result = boto_iot.list_topic_rules(**conn_parameters)

        self.assertTrue(result['rules'])

    def test_that_when_listing_topic_rules_fails_the_list_topic_rules_method_returns_error(self):
        '''
        tests False policy error.
        '''
        self.conn.list_topic_rules.side_effect = ClientError(error_content, 'list_topic_rules')
        result = boto_iot.list_topic_rules(**conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('list_topic_rules'))


if __name__ == '__main__':
    from integration import run_tests  # pylint: disable=import-error
    run_tests(BotoIoTPolicyTestCase, needs_daemon=False)
    run_tests(BotoIoTTopicRuleTestCase, needs_daemon=False)



# -*- coding: utf-8 -*-,
'''
Dell FX2 chassis
================

.. versionadded:: 2015.8.2

Proxy minion interface module for managing Dell FX2 chassis (Dell
Chassis Management Controller version 1.2 and above, iDRAC8 version 2.00
and above)

Dependencies
------------

- :mod:`iDRAC Remote execution module (salt.modules.dracr) <salt.modules.dracr>`
- :mod:`Chassis command shim (salt.modules.chassis) <salt.modules.chassis>`
- :mod:`Dell Chassis States (salt.states.dellchassis) <salt.states.dellchassis>`
- Dell's ``racadm`` command line interface to CMC and iDRAC devices.


**Special Note: SaltStack thanks** `Adobe Corporation <http://adobe.com/>`_
**for their support in creating this proxy minion integration.**

This proxy minion enables Dell FX2 and FX2s (hereafter referred to as
simply "chassis", "CMC", or "FX2") chassis to be treated individually
like a salt-minion.

Since the CMC embedded in the chassis does not run an OS capable of hosting a
Python stack, the chassis can't run a minion directly.  Salt's "Proxy Minion"
functionality enables you to designate another machine to host a minion
process that "proxies" communication from the salt-master.  The master does not
know nor care that the target is not a real minion.

More in-depth conceptual reading on Proxy Minions can be found
:ref:`in the Proxy Minion section <proxy-minion>` of
Salt's documentation.

To configure this integration, follow these steps:

Pillar
------

Proxy minions get their configuration from Salt's Pillar.  Every proxy must
have a stanza in Pillar, and a reference in the Pillar topfile that matches
the ID.  At a minimum for communication with the chassis the pillar should
look like this:

.. code-block:: yaml

    proxy:
      host: <ip or dns name of chassis controller>
      admin_username: <iDRAC username for the CMC, usually 'root'>
      fallback_admin_username: <username to try if the first fails>
      passwords:
        - first_password
        - second_password
        - third-password
      proxytype: fx2

The ``proxytype`` line above is critical, it tells Salt which interface to load
from the ``proxy`` directory in Salt's install hierarchy, or from ``/srv/salt/_proxy``
on the salt-master (if you have created your own proxy module, for example).

The proxy integration will try the passwords listed in order.  It is
configured this way so you can have a regular password, a potential
fallback password, and the third password can be the one you intend
to change the chassis to use.  This way, after it is changed, you
should not need to restart the proxy minion--it should just pick up the
third password in the list.  You can then change pillar at will to
move that password to the front and retire the unused ones.

Beware, many Dell CMC and iDRAC units are configured to lockout
IP addresses or users after too many failed password attempts.  This can
generate user panic in the form of "I no longer know what the password is!!!".
To mitigate panic try the web interface from a different IP, or setup a
emergency administrator user in the CMC before doing a wholesale
password rotation.

The automatic lockout can be disabled via Salt with the following:

.. code-block:: bash

    salt <cmc> chassis.cmd set_general cfgRacTuning cfgRacTuneIpBlkEnable 0

and then verified with

.. code-block:: bash

    salt <cmc> chassis.cmd get_general cfgRacTuning cfgRacTuneIpBlkEnable



salt-proxy
----------

After your pillar is in place, you can test the proxy.  The proxy can run on
any machine that has network connectivity to your salt-master and to the chassis in question.
SaltStack recommends that this machine also run a regular minion, though
it is not strictly necessary.

On the machine that will run the proxy, make sure there is an ``/etc/salt/proxy``
file with at least the following in it:

.. code-block:: yaml

    master: <ip or hostname of salt-master>

You can start the proxy with

.. code-block:: bash

    salt-proxy --proxyid <id you want to give the chassis>

You may want to add ``-l debug`` to run the above in the foreground in debug
mode just to make sure everything is OK.

Next, accept the key for the proxy on your salt-master, just like you would
for a regular minion:

.. code-block:: bash

    salt-key -a <id you want to give the chassis>

You can confirm that the pillar data is in place for the proxy:

.. code-block:: bash

    salt <id> pillar.items

And now you should be able to ping the chassis to make sure it is responding:

.. code-block:: bash

    salt <id> test.ping

At this point you can execute one-off commands against the chassis.  For
example, you can get the chassis inventory:

.. code-block:: bash

    salt <id> chassis.cmd inventory

Note that you don't need to provide credentials or an ip/hostname.  Salt knows
to use the credentials you stored in Pillar.

It's important to understand how this particular proxy works.
:mod:`Salt.modules.dracr <salt.modules.dracr>` is a standard Salt execution
module.  If you pull up the docs for it you'll see that almost every function
in the module takes credentials and a target host.  When credentials and a host
aren't passed, Salt runs ``racadm`` against the local machine.  If you wanted
you could run functions from this module on any host where an appropriate
version of ``racadm`` is installed, and that host would reach out over the network
and communicate with the chassis.

``Chassis.cmd`` acts as a "shim" between the execution module and the proxy.  It's
first parameter is always the function from salt.modules.dracr to execute.  If the
function takes more positional or keyword arguments you can append them to the call.
It's this shim that speaks to the chassis through the proxy, arranging for the
credentials and hostname to be pulled from the pillar section for this proxy minion.

Because of the presence of the shim, to lookup documentation for what
functions you can use to interface with the chassis, you'll want to
look in :mod:`salt.modules.dracr <salt.modules.dracr>` instead
of :mod:`salt.modules.chassis <salt.modules.chassis>`.

States
------

Associated states are thoroughly documented in :mod:`salt.states.dellchassis <salt.states.dellchassis>`.
Look there to find an example structure for pillar as well as an example
``.sls`` file for standing up a Dell Chassis from scratch.

'''
from __future__ import absolute_import

# Import python libs
import logging
import salt.utils
import salt.utils.http

# This must be present or the Salt loader won't load this module
__proxyenabled__ = ['fx2']


# Variables are scoped to this module so we can have persistent data
# across calls to fns in here.
GRAINS_CACHE = {}
DETAILS = {}

# Want logging!
log = logging.getLogger(__file__)


def __virtual__():
    '''
    Only return if all the modules are available
    '''
    if not salt.utils.which('racadm'):
        log.critical('fx2 proxy minion needs "racadm" to be installed.')
        return False

    return True


def init(opts):
    '''
    This function gets called when the proxy starts up.
    We check opts to see if a fallback user and password are supplied.
    If they are present, and the primary credentials don't work, then
    we try the backup before failing.

    Whichever set of credentials works is placed in the persistent
    DETAILS dictionary and will be used for further communication with the
    chassis.
    '''
    if 'host' not in opts['proxy']:
        log.critical('No "host" key found in pillar for this proxy')
        return False

    DETAILS['host'] = opts['proxy']['host']

    (username, password) = find_credentials()


def admin_username():
    '''
    Return the admin_username in the DETAILS dictionary, or root if there
    is none present
    '''
    return DETAILS.get('admin_username', 'root')


def admin_password():
    '''
    Return the admin_password in the DETAILS dictionary, or 'calvin'
    (the Dell default) if there is none present
    '''
    if 'admin_password' not in DETAILS:
        log.info('proxy.fx2: No admin_password in DETAILS, returning Dell default')
        return 'calvin'

    return DETAILS.get('admin_password', 'calvin')


def host():
    return DETAILS['host']


def _grains(host, user, password):
    '''
    Get the grains from the proxied device
    '''
    r = __salt__['dracr.system_info'](host=host,
                                      admin_username=user,
                                      admin_password=password)
    if r.get('retcode', 0) == 0:
        GRAINS_CACHE = r
    else:
        GRAINS_CACHE = {}
    return GRAINS_CACHE


def grains():
    '''
    Get the grains from the proxied device
    '''
    if not GRAINS_CACHE:
        return _grains(DETAILS['host'],
                       DETAILS['admin_username'],
                       DETAILS['admin_password'])

    return GRAINS_CACHE


def grains_refresh():
    '''
    Refresh the grains from the proxied device
    '''
    GRAINS_CACHE = {}
    return grains()


def find_credentials():
    '''
    Cycle through all the possible credentials and return the first one that
    works
    '''
    usernames = []
    usernames.append(__pillar__['proxy'].get('admin_username', 'root'))
    if 'fallback_admin_username' in __pillar__.get('proxy'):
        usernames.append(__pillar__['proxy'].get('fallback_admin_username'))

    for user in usernames:
        for pwd in __pillar__['proxy']['passwords']:
            r = __salt__['dracr.get_chassis_name'](host=__pillar__['proxy']['host'],
                                                   admin_username=user,
                                                   admin_password=pwd)
            # Retcode will be present if the chassis_name call failed
            try:
                if r.get('retcode', None) is None:
                    DETAILS['admin_username'] = user
                    DETAILS['admin_password'] = pwd
                    __opts__['proxy']['admin_username'] = user
                    __opts__['proxy']['admin_password'] = pwd
                    return (user, pwd)
            except AttributeError:
                # Then the above was a string, and we can return the username
                # and password
                DETAILS['admin_username'] = user
                DETAILS['admin_password'] = pwd
                __opts__['proxy']['admin_username'] = user
                __opts__['proxy']['admin_password'] = pwd
                return (user, pwd)

    log.debug('proxy fx2.find_credentials found no valid credentials, using Dell default')
    return ('root', 'calvin')


def chconfig(cmd, *args, **kwargs):
    '''
    This function is called by the :mod:`salt.modules.chassis.cmd <salt.modules.chassis.cmd>`
    shim.  It then calls whatever is passed in ``cmd``
    inside the :mod:`salt.modules.dracr <salt.modules.dracr>`
    module.

    :param cmd: The command to call inside salt.modules.dracr
    :param args: Arguments that need to be passed to that command
    :param kwargs: Keyword arguments that need to be passed to that command
    :return: Passthrough the return from the dracr module.

    '''
    # Strip the __pub_ keys...is there a better way to do this?
    for k in kwargs.keys():
        if k.startswith('__pub_'):
            kwargs.pop(k)

    # Catch password reset
    if 'dracr.'+cmd not in __salt__:
        ret = {'retcode': -1, 'message': 'dracr.' + cmd + ' is not available'}
    else:
        ret = __salt__['dracr.'+cmd](*args, **kwargs)

    if cmd == 'change_password':
        if 'username' in kwargs:
            __opts__['proxy']['admin_username'] = kwargs['username']
            DETAILS['admin_username'] = kwargs['username']
        if 'password' in kwargs:
            __opts__['proxy']['admin_password'] = kwargs['password']
            DETAILS['admin_password'] = kwargs['password']

    return ret


def ping():
    '''
    Is the chassis responding?

    :return: Returns False if the chassis didn't respond, True otherwise.

    '''
    r = __salt__['dracr.system_info'](host=DETAILS['host'],
                                      admin_username=DETAILS['admin_username'],
                                      admin_password=DETAILS['admin_password'])
    if r.get('retcode', 0) == 1:
        return False
    else:
        return True
    try:
        return r['dict'].get('ret', False)
    except Exception:
        return False


def shutdown(opts):
    '''
    Shutdown the connection to the proxied device.
    For this proxy shutdown is a no-op.
    '''
    log.debug('fx2 proxy shutdown() called...')



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import

# Import Salt Libs
from salt.states import win_certutil as certutil

# Import Salt Testing Libs
from salttesting import TestCase
from salttesting.helpers import ensure_in_syspath
from salttesting.mock import (
    MagicMock,
    patch
)

ensure_in_syspath('../../')

certutil.__salt__ = {}


class CertUtilTestCase(TestCase):

    def test_add_serial(self):
        '''
            Test adding a certificate to specified certificate store
        '''
        expected = {
            'changes': {'added': '/path/to/cert.cer'},
            'comment': '',
            'name': '/path/to/cert.cer',
            'result': True
        }

        cache_mock = MagicMock(return_value='/tmp/cert.cer')
        get_cert_serial_mock = MagicMock(return_value='ABCDEF')
        get_store_serials_mock = MagicMock(return_value=['123456'])
        add_mock = MagicMock(return_value='Added successfully')
        with patch.dict(certutil.__salt__, {'cp.cache_file': cache_mock,
                                            'certutil.get_cert_serial': get_cert_serial_mock,
                                            'certutil.get_stored_cert_serials': get_store_serials_mock,
                                            'certutil.add_store': add_mock}):
            out = certutil.add_store('/path/to/cert.cer', 'TrustedPublisher')
            cache_mock.assert_called_once_with('/path/to/cert.cer', 'base')
            get_cert_serial_mock.assert_called_once_with('/tmp/cert.cer')
            get_store_serials_mock.assert_called_once_with('TrustedPublisher')
            add_mock.assert_called_once_with('/path/to/cert.cer', 'TrustedPublisher')
            self.assertEqual(expected, out)

    def test_add_serial_missing(self):
        '''
            Test adding a certificate to specified certificate store when the file doesn't exist
        '''
        expected = {
            'changes': {},
            'comment': 'Certificate file not found.',
            'name': '/path/to/cert.cer',
            'result': False
        }

        cache_mock = MagicMock(return_value=False)
        get_cert_serial_mock = MagicMock(return_value='ABCDEF')
        get_store_serials_mock = MagicMock(return_value=['123456'])
        add_mock = MagicMock(return_value='Added successfully')
        with patch.dict(certutil.__salt__, {'cp.cache_file': cache_mock,
                                            'certutil.get_cert_serial': get_cert_serial_mock,
                                            'certutil.get_stored_cert_serials': get_store_serials_mock,
                                            'certutil.add_store': add_mock}):
            out = certutil.add_store('/path/to/cert.cer', 'TrustedPublisher')
            cache_mock.assert_called_once_with('/path/to/cert.cer', 'base')
            assert not get_cert_serial_mock.called
            assert not get_store_serials_mock.called
            assert not add_mock.called
            self.assertEqual(expected, out)

    def test_add_serial_exists(self):
        '''
            Test adding a certificate to specified certificate store when the cert already exists
        '''
        expected = {
            'changes': {},
            'comment': '/path/to/cert.cer already stored.',
            'name': '/path/to/cert.cer',
            'result': True
        }

        cache_mock = MagicMock(return_value='/tmp/cert.cer')
        get_cert_serial_mock = MagicMock(return_value='ABCDEF')
        get_store_serials_mock = MagicMock(return_value=['123456', 'ABCDEF'])
        add_mock = MagicMock(return_value='Added successfully')
        with patch.dict(certutil.__salt__, {'cp.cache_file': cache_mock,
                                            'certutil.get_cert_serial': get_cert_serial_mock,
                                            'certutil.get_stored_cert_serials': get_store_serials_mock,
                                            'certutil.add_store': add_mock}):
            out = certutil.add_store('/path/to/cert.cer', 'TrustedPublisher')
            cache_mock.assert_called_once_with('/path/to/cert.cer', 'base')
            get_cert_serial_mock.assert_called_once_with('/tmp/cert.cer')
            get_store_serials_mock.assert_called_once_with('TrustedPublisher')
            assert not add_mock.called
            self.assertEqual(expected, out)

    def test_add_serial_fail(self):
        '''
            Test adding a certificate when the add fails
        '''
        expected = {
            'changes': {},
            'comment': 'Failed to store certificate /path/to/cert.cer',
            'name': '/path/to/cert.cer',
            'result': False
        }

        cache_mock = MagicMock(return_value='/tmp/cert.cer')
        get_cert_serial_mock = MagicMock(return_value='ABCDEF')
        get_store_serials_mock = MagicMock(return_value=['123456'])
        add_mock = MagicMock(return_value='Failed')
        with patch.dict(certutil.__salt__, {'cp.cache_file': cache_mock,
                                            'certutil.get_cert_serial': get_cert_serial_mock,
                                            'certutil.get_stored_cert_serials': get_store_serials_mock,
                                            'certutil.add_store': add_mock}):
            out = certutil.add_store('/path/to/cert.cer', 'TrustedPublisher')
            cache_mock.assert_called_once_with('/path/to/cert.cer', 'base')
            get_cert_serial_mock.assert_called_once_with('/tmp/cert.cer')
            get_store_serials_mock.assert_called_once_with('TrustedPublisher')
            add_mock.assert_called_once_with('/path/to/cert.cer', 'TrustedPublisher')
            self.assertEqual(expected, out)

    def test_del_serial(self):
        '''
            Test deleting a certificate from a specified certificate store
        '''
        expected = {
            'changes': {'removed': '/path/to/cert.cer'},
            'comment': '',
            'name': '/path/to/cert.cer',
            'result': True
        }

        cache_mock = MagicMock(return_value='/tmp/cert.cer')
        get_cert_serial_mock = MagicMock(return_value='ABCDEF')
        get_store_serials_mock = MagicMock(return_value=['123456', 'ABCDEF'])
        del_mock = MagicMock(return_value='Removed successfully')
        with patch.dict(certutil.__salt__, {'cp.cache_file': cache_mock,
                                            'certutil.get_cert_serial': get_cert_serial_mock,
                                            'certutil.get_stored_cert_serials': get_store_serials_mock,
                                            'certutil.del_store': del_mock}):
            out = certutil.del_store('/path/to/cert.cer', 'TrustedPublisher')
            cache_mock.assert_called_once_with('/path/to/cert.cer', 'base')
            get_cert_serial_mock.assert_called_once_with('/tmp/cert.cer')
            get_store_serials_mock.assert_called_once_with('TrustedPublisher')
            del_mock.assert_called_once_with('/tmp/cert.cer', 'TrustedPublisher')
            self.assertEqual(expected, out)

    def test_del_serial_missing(self):
        '''
            Test deleting a certificate to specified certificate store when the file doesn't exist
        '''
        expected = {
            'changes': {},
            'comment': 'Certificate file not found.',
            'name': '/path/to/cert.cer',
            'result': False
        }

        cache_mock = MagicMock(return_value=False)
        get_cert_serial_mock = MagicMock(return_value='ABCDEF')
        get_store_serials_mock = MagicMock(return_value=['123456'])
        del_mock = MagicMock(return_value='Added successfully')
        with patch.dict(certutil.__salt__, {'cp.cache_file': cache_mock,
                                            'certutil.get_cert_serial': get_cert_serial_mock,
                                            'certutil.get_stored_cert_serials': get_store_serials_mock,
                                            'certutil.del_store': del_mock}):
            out = certutil.del_store('/path/to/cert.cer', 'TrustedPublisher')
            cache_mock.assert_called_once_with('/path/to/cert.cer', 'base')
            assert not get_cert_serial_mock.called
            assert not get_store_serials_mock.called
            assert not del_mock.called
            self.assertEqual(expected, out)

    def test_del_serial_doesnt_exists(self):
        '''
            Test deleting a certificate to specified certificate store when the cert doesn't exists
        '''
        expected = {
            'changes': {},
            'comment': '/path/to/cert.cer already removed.',
            'name': '/path/to/cert.cer',
            'result': True
        }

        cache_mock = MagicMock(return_value='/tmp/cert.cer')
        get_cert_serial_mock = MagicMock(return_value='ABCDEF')
        get_store_serials_mock = MagicMock(return_value=['123456'])
        del_mock = MagicMock(return_value='Added successfully')
        with patch.dict(certutil.__salt__, {'cp.cache_file': cache_mock,
                                            'certutil.get_cert_serial': get_cert_serial_mock,
                                            'certutil.get_stored_cert_serials': get_store_serials_mock,
                                            'certutil.del_store': del_mock}):
            out = certutil.del_store('/path/to/cert.cer', 'TrustedPublisher')
            cache_mock.assert_called_once_with('/path/to/cert.cer', 'base')
            get_cert_serial_mock.assert_called_once_with('/tmp/cert.cer')
            get_store_serials_mock.assert_called_once_with('TrustedPublisher')
            assert not del_mock.called
            self.assertEqual(expected, out)

    def test_del_serial_fail(self):
        '''
            Test deleting a certificate from the store when the delete fails
        '''
        expected = {
            'changes': {},
            'comment': 'Failed to remove the certificate /path/to/cert.cer',
            'name': '/path/to/cert.cer',
            'result': False
        }

        cache_mock = MagicMock(return_value='/tmp/cert.cer')
        get_cert_serial_mock = MagicMock(return_value='ABCDEF')
        get_store_serials_mock = MagicMock(return_value=['123456', 'ABCDEF'])
        del_mock = MagicMock(return_value='Failed')
        with patch.dict(certutil.__salt__, {'cp.cache_file': cache_mock,
                                            'certutil.get_cert_serial': get_cert_serial_mock,
                                            'certutil.get_stored_cert_serials': get_store_serials_mock,
                                            'certutil.del_store': del_mock}):
            out = certutil.del_store('/path/to/cert.cer', 'TrustedPublisher')
            cache_mock.assert_called_once_with('/path/to/cert.cer', 'base')
            get_cert_serial_mock.assert_called_once_with('/tmp/cert.cer')
            get_store_serials_mock.assert_called_once_with('TrustedPublisher')
            del_mock.assert_called_once_with('/tmp/cert.cer', 'TrustedPublisher')
            self.assertEqual(expected, out)


if __name__ == '__main__':
    from integration import run_tests
    run_tests(CertUtilTestCase, needs_daemon=False)



# -*- coding: utf-8 -*-
'''
unit tests for the grains state
'''

from __future__ import absolute_import

# Import Python libs
import os
import yaml

# Import Salt Testing libs
from salttesting import TestCase, skipIf
from salttesting.helpers import ensure_in_syspath
from salttesting.mock import NO_MOCK, NO_MOCK_REASON, MagicMock, patch

ensure_in_syspath('../../')

from salt.modules import grains as grainsmod
from salt.states import grains as grains

import integration


@skipIf(NO_MOCK, NO_MOCK_REASON)
class GrainsTestCase(TestCase):

    def setUp(self):
        grains_test_dir = '__salt_test_state_grains'
        grainsmod.__opts__ = grains.__opts__ = {
            'test': False,
            'conf_file': os.path.join(integration.TMP, grains_test_dir, 'minion'),
            'cachedir':  os.path.join(integration.TMP, grains_test_dir),
            'local': True,
        }

        grainsmod.__salt__ = grains.__salt__ = {
            'cmd.run_all': MagicMock(return_value={
                'pid': 5,
                'retcode': 0,
                'stderr': '',
                'stdout': ''}),
            'grains.get':    grainsmod.get,
            'grains.set':    grainsmod.set,
            'grains.setval': grainsmod.setval,
            'grains.delval': grainsmod.delval,
            'grains.append': grainsmod.append,
            'grains.remove': grainsmod.remove,
            'saltutil.sync_grains': MagicMock()
        }
        if not os.path.exists(os.path.join(integration.TMP, grains_test_dir)):
            os.mkdir(os.path.join(integration.TMP, grains_test_dir))

    def assertGrainFileContent(self, grains_string):
        if os.path.isdir(grains.__opts__['conf_file']):
            grains_file = os.path.join(
                grains.__opts__['conf_file'],
                'grains')
        else:
            grains_file = os.path.join(
                os.path.dirname(grains.__opts__['conf_file']),
                'grains')
        with open(grains_file, "r") as grf:
            grains_data = grf.read()
        self.assertMultiLineEqual(grains_string, grains_data)

    def setGrains(self, grains_data):
        grains.__grains__ = grainsmod.__grains__ = grains_data
        if os.path.isdir(grains.__opts__['conf_file']):
            grains_file = os.path.join(
                grains.__opts__['conf_file'],
                'grains')
        else:
            grains_file = os.path.join(
                os.path.dirname(grains.__opts__['conf_file']),
                'grains')
        cstr = yaml.safe_dump(grains_data, default_flow_style=False)
        with open(grains_file, "w+") as grf:
            grf.write(cstr)

    # 'present' function tests: 12

    def test_present_add(self):
        # Set a non existing grain
        self.setGrains({'a': 'aval'})
        ret = grains.present(
            name='foo',
            value='bar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['changes'], {'foo': 'bar'})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': 'bar'})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: bar\n"
        )

        # Set a non existing nested grain
        self.setGrains({'a': 'aval'})
        ret = grains.present(
            name='foo:is:nested',
            value='bar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['changes'], {'foo': {'is': {'nested': 'bar'}}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  is:\n"
                                  + "    nested: bar\n"
        )

        # Set a non existing nested dict grain
        self.setGrains({'a': 'aval'})
        ret = grains.present(
            name='foo:is:nested',
            value={'bar': 'is a dict'})
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['changes'], {'foo': {'is': {'nested': {'bar': 'is a dict'}}}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': {'bar': 'is a dict'}}}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  is:\n"
                                  + "    nested:\n"
                                  + "      bar: is a dict\n"
        )

    def test_present_add_key_to_existing(self):
        self.setGrains({'a': 'aval', 'foo': {'k1': 'v1'}})
        # Fails setting a grain to a dict
        ret = grains.present(
            name='foo:k2',
            value='v2')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Set grain foo:k2 to v2')
        self.assertEqual(ret['changes'], {'foo': {'k2': 'v2', 'k1': 'v1'}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'k1': 'v1', 'k2': 'v2'}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  k1: v1\n"
                                  + "  k2: v2\n"
        )

    def test_present_already_set(self):
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Grain already set
        ret = grains.present(
            name='foo',
            value='bar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain is already set')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': 'bar'})

        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})
        # Nested grain already set
        ret = grains.present(
            name='foo:is:nested',
            value='bar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain is already set')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})

        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})
        # Nested dict grain already set
        ret = grains.present(
            name='foo:is',
            value={'nested': 'bar'})
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain is already set')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})

    def test_present_overwrite(self):
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Overwrite an existing grain
        ret = grains.present(
            name='foo',
            value='newbar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['changes'], {'foo': 'newbar'})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': 'newbar'})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: newbar\n"
        )

        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Clear a grain (set to None)
        ret = grains.present(
            name='foo',
            value=None)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['changes'], {'foo': None})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': None})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: null\n"
        )

        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})
        # Overwrite an existing nested grain
        ret = grains.present(
            name='foo:is:nested',
            value='newbar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['changes'], {'foo': {'is': {'nested': 'newbar'}}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': 'newbar'}}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  is:\n"
                                  + "    nested: newbar\n"
        )

        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})
        # Clear a nested grain (set to None)
        ret = grains.present(
            name='foo:is:nested',
            value=None)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['changes'], {'foo': {'is': {'nested': None}}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': None}}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  is:\n"
                                  + "    nested: null\n"
        )

    def test_present_fail_overwrite(self):
        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'val'}}})
        # Overwrite an existing grain
        ret = grains.present(
            name='foo:is',
            value='newbar')
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['changes'], {})
        self.assertEqual(ret['comment'], 'The key \'foo:is\' exists but is a dict or a list. Use \'force=True\' to overwrite.')
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': 'val'}}})

        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'val'}}})
        # Clear a grain (set to None)
        ret = grains.present(
            name='foo:is',
            value=None)
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['changes'], {})
        self.assertEqual(ret['comment'], 'The key \'foo:is\' exists but is a dict or a list. Use \'force=True\' to overwrite.')
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': 'val'}}})

    def test_present_fails_to_set_dict_or_list(self):
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Fails to overwrite a grain to a list
        ret = grains.present(
            name='foo',
            value=['l1', 'l2'])
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'The key \'foo\' exists and the '
                                       + 'given value is a dict or a list. '
                                       + 'Use \'force=True\' to overwrite.')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': 'bar'})

        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Fails setting a grain to a dict
        ret = grains.present(
            name='foo',
            value={'k1': 'v1'})
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'The key \'foo\' exists and the given '
                                       + 'value is a dict or a list. Use '
                                       + '\'force=True\' to overwrite.')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': 'bar'})

        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})
        # Fails to overwrite a nested grain to a list
        ret = grains.present(
            name='foo,is,nested',
            value=['l1', 'l2'],
            delimiter=',')
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['changes'], {})
        self.assertEqual(ret['comment'], 'The key \'foo:is:nested\' exists and the '
                                       + 'given value is a dict or a list. '
                                       + 'Use \'force=True\' to overwrite.')
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})

        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})
        # Fails setting a nested grain to a dict
        ret = grains.present(
            name='foo:is:nested',
            value={'k1': 'v1'})
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'The key \'foo:is:nested\' exists and the '
                                       + 'given value is a dict or a list. '
                                       + 'Use \'force=True\' to overwrite.')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})

    def test_present_fail_merge_dict(self):
        self.setGrains({'a': 'aval', 'foo': {'k1': 'v1'}})
        # Fails setting a grain to a dict
        ret = grains.present(
            name='foo',
            value={'k2': 'v2'})
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'The key \'foo\' exists but '
                                       + 'is a dict or a list. '
                                       + 'Use \'force=True\' to overwrite.')
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'k1': 'v1'}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  k1: v1\n"
        )

    def test_present_force_to_set_dict_or_list(self):
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Force to overwrite a grain to a list
        ret = grains.present(
            name='foo',
            value=['l1', 'l2'],
            force=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Set grain foo to [\'l1\', \'l2\']')
        self.assertEqual(ret['changes'], {'foo': ['l1', 'l2']})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['l1', 'l2']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- l1\n"
                                  + "- l2\n"
        )

        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Force setting a grain to a dict
        ret = grains.present(
            name='foo',
            value={'k1': 'v1'},
            force=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Set grain foo to {\'k1\': \'v1\'}')
        self.assertEqual(ret['changes'], {'foo': {'k1': 'v1'}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'k1': 'v1'}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  k1: v1\n"
        )

        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})
        # Force to overwrite a nested grain to a list
        ret = grains.present(
            name='foo,is,nested',
            value=['l1', 'l2'],
            delimiter=',',
            force=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['changes'], {'foo': {'is': {'nested': ['l1', 'l2']}}})
        self.assertEqual(ret['comment'], 'Set grain foo:is:nested to [\'l1\', \'l2\']')
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': ['l1', 'l2']}}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  is:\n"
                                  + "    nested:\n"
                                  + "    - l1\n"
                                  + "    - l2\n"
        )

        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}, 'and': 'other'}})
        # Force setting a nested grain to a dict
        ret = grains.present(
            name='foo:is:nested',
            value={'k1': 'v1'},
            force=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Set grain foo:is:nested to {\'k1\': \'v1\'}')
        self.assertEqual(ret['changes'], {'foo': {'is': {'nested': {'k1': 'v1'}}, 'and': 'other'}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': {'k1': 'v1'}}, 'and': 'other'}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  and: other\n"
                                  + "  is:\n"
                                  + "    nested:\n"
                                  + "      k1: v1\n"
        )

    def test_present_fails_to_convert_value_to_key(self):
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Fails converting a value to a nested grain key
        ret = grains.present(
            name='foo:is:nested',
            value={'k1': 'v1'})
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'The key \'foo\' value is \'bar\', '
                               + 'which is different from the provided '
                               + 'key \'is\'. Use \'force=True\' to overwrite.')
        self.assertEqual(ret['changes'], {})

    def test_present_overwrite_test(self):
        grains.__opts__['test'] = True
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Overwrite an existing grain
        ret = grains.present(
            name='foo',
            value='newbar')
        self.assertEqual(ret['result'], None)
        self.assertEqual(ret['changes'], {'changed': {'foo': 'newbar'}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': 'bar'})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: bar\n"
        )

    def test_present_convert_value_to_key(self):
        self.setGrains({'a': 'aval', 'foo': 'is'})
        # Converts a value to a nested grain key
        ret = grains.present(
            name='foo:is:nested',
            value={'k1': 'v1'})
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Set grain foo:is:nested to {\'k1\': \'v1\'}')
        self.assertEqual(ret['changes'], {'foo': {'is': {'nested': {'k1': 'v1'}}}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': {'k1': 'v1'}}}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  is:\n"
                                  + "    nested:\n"
                                  + "      k1: v1\n"
        )

        self.setGrains({'a': 'aval', 'foo': ['one', 'is', 'correct']})
        # Converts a list element to a nested grain key
        ret = grains.present(
            name='foo:is:nested',
            value={'k1': 'v1'})
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Set grain foo:is:nested to {\'k1\': \'v1\'}')
        self.assertEqual(ret['changes'], {'foo': ['one', {'is': {'nested': {'k1': 'v1'}}}, 'correct']})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['one', {'is': {'nested': {'k1': 'v1'}}}, 'correct']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- one\n"
                                  + "- is:\n"
                                  + "    nested:\n"
                                  + "      k1: v1\n"
                                  + "- correct\n"
        )

    @patch('salt.modules.grains.setval')
    def test_present_unknown_failure(self, mocked_setval):
        mocked_setval.return_value = 'Failed to set grain foo'
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Unknown reason failure
        ret = grains.present(
            name='foo',
            value='baz')
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'Failed to set grain foo')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': 'bar'})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: bar\n"
        )

    # 'absent' function tests: 6

    def test_absent_already(self):
        # Unset a non existent grain
        self.setGrains({'a': 'aval'})
        ret = grains.absent(
            name='foo')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain foo does not exist')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval'})
        self.assertGrainFileContent("a: aval\n")

        # Unset a non existent nested grain
        self.setGrains({'a': 'aval'})
        ret = grains.absent(
            name='foo:is:nested')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain foo:is:nested does not exist')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval'})
        self.assertGrainFileContent("a: aval\n")

    def test_absent_unset(self):
        # Unset a grain
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        ret = grains.absent(
            name='foo')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value for grain foo was set to None')
        self.assertEqual(ret['changes'], {'grain': 'foo', 'value': None})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': None})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: null\n"
        )

        # Unset grain when its value is False
        self.setGrains({'a': 'aval', 'foo': False})
        ret = grains.absent(
            name='foo')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value for grain foo was set to None')
        self.assertEqual(ret['changes'], {'grain': 'foo', 'value': None})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': None})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: null\n"
        )

        # Unset a nested grain
        self.setGrains({'a': 'aval', 'foo': ['order', {'is': {'nested': 'bar'}}, 'correct']})
        ret = grains.absent(
            name='foo,is,nested',
            delimiter=',')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value for grain foo:is:nested was set to None')
        self.assertEqual(ret['changes'], {'grain': 'foo:is:nested', 'value': None})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['order', {'is': {'nested': None}}, 'correct']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- order\n"
                                  + "- is:\n"
                                  + "    nested: null\n"
                                  + "- correct\n"
        )

        # Unset a nested value don't change anything
        self.setGrains({'a': 'aval', 'foo': ['order', {'is': 'nested'}, 'correct']})
        ret = grains.absent(
            name='foo:is:nested')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain foo:is:nested does not exist')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['order', {'is': 'nested'}, 'correct']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- order\n"
                                  + "- is: nested\n"
                                  + "- correct\n"
        )

    def test_absent_unset_test(self):
        grains.__opts__['test'] = True
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        # Overwrite an existing grain
        ret = grains.absent(name='foo')
        self.assertEqual(ret['result'], None)
        self.assertEqual(ret['changes'], {'grain': 'foo', 'value': None})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': 'bar'})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: bar\n"
        )

    def test_absent_fails_nested_complex_grain(self):
        # Unset a nested complex grain
        self.setGrains({'a': 'aval', 'foo': ['order', {'is': {'nested': 'bar'}}, 'correct']})
        ret = grains.absent(
            name='foo:is')
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'The key \'foo:is\' exists but is a dict or a list. Use \'force=True\' to overwrite.')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['order', {'is': {'nested': 'bar'}}, 'correct']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- order\n"
                                  + "- is:\n"
                                  + "    nested: bar\n"
                                  + "- correct\n"
        )

    def test_absent_force_nested_complex_grain(self):
        # Unset a nested complex grain
        self.setGrains({'a': 'aval', 'foo': ['order', {'is': {'nested': 'bar'}}, 'correct']})
        ret = grains.absent(
            name='foo:is',
            force=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value for grain foo:is was set to None')
        self.assertEqual(ret['changes'], {'grain': 'foo:is', 'value': None})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['order', {'is': None}, 'correct']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- order\n"
                                  + "- is: null\n"
                                  + "- correct\n"
        )

    def test_absent_delete(self):
        # Delete a grain
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        ret = grains.absent(
            name='foo',
            destructive=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain foo was deleted')
        self.assertEqual(ret['changes'], {'deleted': 'foo'})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval'})
        self.assertGrainFileContent("a: aval\n")

        # Delete a previously unset grain
        self.setGrains({'a': 'aval', 'foo': None})
        ret = grains.absent(
            name='foo',
            destructive=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain foo was deleted')
        self.assertEqual(ret['changes'], {'deleted': 'foo'})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval'})
        self.assertGrainFileContent("a: aval\n")

        # Delete a nested grain
        self.setGrains({'a': 'aval', 'foo': ['order', {'is': {'nested': 'bar', 'other': 'value'}}, 'correct']})
        ret = grains.absent(
            name='foo:is:nested',
            destructive=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain foo:is:nested was deleted')
        self.assertEqual(ret['changes'], {'deleted': 'foo:is:nested'})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['order', {'is': {'other': 'value'}}, 'correct']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- order\n"
                                  + "- is:\n"
                                  + "    other: value\n"
                                  + "- correct\n"
        )

    # 'append' function tests: 6

    def test_append(self):
        # Append to an existing list
        self.setGrains({'a': 'aval', 'foo': ['bar']})
        ret = grains.append(
            name='foo',
            value='baz')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value baz was added to grain foo')
        self.assertEqual(ret['changes'], {'added': 'baz'})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['bar', 'baz']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- bar\n"
                                  + "- baz\n"
        )

    def test_append_nested(self):
        # Append to an existing nested list
        self.setGrains({'a': 'aval', 'foo': {'list': ['bar']}})
        ret = grains.append(
            name='foo,list',
            value='baz',
            delimiter=',')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value baz was added to grain foo:list')
        self.assertEqual(ret['changes'], {'added': 'baz'})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'list': ['bar', 'baz']}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  list:\n"
                                  + "  - bar\n"
                                  + "  - baz\n"
        )

    def test_append_already(self):
        # Append to an existing list
        self.setGrains({'a': 'aval', 'foo': ['bar']})
        ret = grains.append(
            name='foo',
            value='bar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value bar is already in the list '
                                       + 'for grain foo')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['bar']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- bar\n"
        )

    def test_append_fails_not_a_list(self):
        # Fail to append to an existing grain, not a list
        self.setGrains({'a': 'aval', 'foo': {'bar': 'val'}})
        ret = grains.append(
            name='foo',
            value='baz')
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'Grain foo is not a valid list')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'bar': 'val'}})

    def test_append_convert_to_list(self):
        # Append to an existing grain, converting to a list
        self.setGrains({'a': 'aval', 'foo': {'bar': 'val'}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  bar: val\n"
        )
        ret = grains.append(
            name='foo',
            value='baz',
            convert=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value baz was added to grain foo')
        self.assertEqual(ret['changes'], {'added': 'baz'})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': [{'bar': 'val'}, 'baz']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- bar: val\n"
                                  + "- baz\n"
        )

        # Append to an existing grain, converting to a list a multi-value dict
        self.setGrains({'a': 'aval', 'foo': {'bar': 'val', 'other': 'value'}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  bar: val\n"
                                  + "  other: value\n"
        )
        ret = grains.append(
            name='foo',
            value='baz',
            convert=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value baz was added to grain foo')
        self.assertEqual(ret['changes'], {'added': 'baz'})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': [{'bar': 'val', 'other': 'value'}, 'baz']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- bar: val\n"
                                  + "  other: value\n"
                                  + "- baz\n"
        )

    def test_append_fails_inexistent(self):
        # Append to a non existing grain
        self.setGrains({'a': 'aval'})
        ret = grains.append(
            name='foo',
            value='bar')
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'Grain foo does not exist')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval'})

    def test_append_convert_to_list_empty(self):
        # Append to an existing list
        self.setGrains({'foo': None})
        ret = grains.append(
            name='foo',
            value='baz',
            convert=True)
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value baz was added to grain foo')
        self.assertEqual(ret['changes'], {'added': 'baz'})
        self.assertEqual(
            grains.__grains__,
            {'foo': ['baz']})
        self.assertGrainFileContent("foo:\n"
                                  + "- baz\n"
        )

    # 'list_present' function tests: 7

    def test_list_present(self):
        self.setGrains({'a': 'aval', 'foo': ['bar']})
        ret = grains.list_present(
            name='foo',
            value='baz')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Append value baz to grain foo')
        self.assertEqual(ret['changes'], {'new': {'foo': ['bar', 'baz']}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['bar', 'baz']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- bar\n"
                                  + "- baz\n"
        )

    def test_list_present_nested(self):
        self.setGrains({'a': 'aval', 'foo': {'is': {'nested': ['bar']}}})
        ret = grains.list_present(
            name='foo,is,nested',
            value='baz',
            delimiter=',')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Append value baz to grain foo:is:nested')
        self.assertEqual(ret['changes'], {'new': {'foo': {'is': {'nested': ['bar', 'baz']}}}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': ['bar', 'baz']}}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  is:\n"
                                  + "    nested:\n"
                                  + "    - bar\n"
                                  + "    - baz\n"
        )

    def test_list_present_inexistent(self):
        self.setGrains({'a': 'aval'})
        ret = grains.list_present(
            name='foo',
            value='baz')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Append value baz to grain foo')
        self.assertEqual(ret['changes'], {'new': {'foo': ['baz']}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['baz']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- baz\n"
        )

    def test_list_present_inexistent_nested(self):
        self.setGrains({'a': 'aval'})
        ret = grains.list_present(
            name='foo:is:nested',
            value='baz')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Append value baz to grain foo:is:nested')
        self.assertEqual(ret['changes'], {'new': {'foo': {'is': {'nested': ['baz']}}}})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'is': {'nested': ['baz']}}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  is:\n"
                                  + "    nested:\n"
                                  + "    - baz\n"
        )

    def test_list_present_not_a_list(self):
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        ret = grains.list_present(
            name='foo',
            value='baz')
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'Grain foo is not a valid list')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': 'bar'})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: bar\n"
        )

    def test_list_present_nested_already(self):
        self.setGrains({'a': 'aval', 'b': {'foo': ['bar']}})
        ret = grains.list_present(
            name='b:foo',
            value='bar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value bar is already in grain b:foo')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'b': {'foo': ['bar']}})
        self.assertGrainFileContent("a: aval\n"
                                  + "b:\n"
                                  + "  foo:\n"
                                  + "  - bar\n"
        )

    def test_list_present_already(self):
        self.setGrains({'a': 'aval', 'foo': ['bar']})
        ret = grains.list_present(
            name='foo',
            value='bar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value bar is already in grain foo')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['bar']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- bar\n"
        )

    def test_list_present_unknown_failure(self):
        self.setGrains({'a': 'aval', 'foo': ['bar']})
        # Unknown reason failure
        grainsmod.__salt__['grains.append'] = MagicMock()
        ret = grains.list_present(
            name='foo',
            value='baz')
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'Failed append value baz to grain foo')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['bar']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- bar\n"
        )

    # 'list_absent' function tests: 6

    def test_list_absent(self):
        self.setGrains({'a': 'aval', 'foo': ['bar']})
        ret = grains.list_absent(
            name='foo',
            value='bar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value bar was deleted from grain foo')
        self.assertEqual(ret['changes'], {'deleted': ['bar']})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': []})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: []\n"
        )

    def test_list_absent_nested(self):
        self.setGrains({'a': 'aval', 'foo': {'list': ['bar']}})
        ret = grains.list_absent(
            name='foo:list',
            value='bar')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value bar was deleted from grain foo:list')
        self.assertEqual(ret['changes'], {'deleted': ['bar']})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': {'list': []}})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "  list: []\n"
        )

    def test_list_absent_inexistent(self):
        self.setGrains({'a': 'aval'})
        ret = grains.list_absent(
            name='foo',
            value='baz')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain foo does not exist')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval'})
        self.assertGrainFileContent("a: aval\n")

    def test_list_absent_inexistent_nested(self):
        self.setGrains({'a': 'aval'})
        ret = grains.list_absent(
            name='foo:list',
            value='baz')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Grain foo:list does not exist')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval'})
        self.assertGrainFileContent("a: aval\n")

    def test_list_absent_not_a_list(self):
        self.setGrains({'a': 'aval', 'foo': 'bar'})
        ret = grains.list_absent(
            name='foo',
            value='bar')
        self.assertEqual(ret['result'], False)
        self.assertEqual(ret['comment'], 'Grain foo is not a valid list')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': 'bar'})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo: bar\n"
        )

    def test_list_absent_already(self):
        self.setGrains({'a': 'aval', 'foo': ['bar']})
        ret = grains.list_absent(
            name='foo',
            value='baz')
        self.assertEqual(ret['result'], True)
        self.assertEqual(ret['comment'], 'Value baz is absent from grain foo')
        self.assertEqual(ret['changes'], {})
        self.assertEqual(
            grains.__grains__,
            {'a': 'aval', 'foo': ['bar']})
        self.assertGrainFileContent("a: aval\n"
                                  + "foo:\n"
                                  + "- bar\n"
        )


if __name__ == '__main__':
    from integration import run_tests
    run_tests(GrainsTestCase, needs_daemon=False)



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import
from distutils.version import LooseVersion  # pylint: disable=import-error,no-name-in-module
import logging
import os
import datetime
import random
import string

# Import Salt Testing libs
from salttesting.unit import skipIf, TestCase
from salttesting.mock import NO_MOCK, NO_MOCK_REASON, MagicMock, patch
from salttesting.helpers import ensure_in_syspath

ensure_in_syspath('../../')

# Import Salt libs
import salt.config
import salt.loader

# Import 3rd-party libs
import yaml

# pylint: disable=import-error,no-name-in-module
from unit.modules.boto_apigateway_test import BotoApiGatewayTestCaseMixin

# Import 3rd-party libs
try:
    import boto3
    from botocore.exceptions import ClientError
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

from salt.ext.six.moves import range

# pylint: enable=import-error,no-name-in-module

# the boto_apigateway module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto3_version = '1.2.1'

region = 'us-east-1'
access_key = 'GKTADJGHEIQSXMKKRBJ08H'
secret_key = 'askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs'
conn_parameters = {'region': region, 'key': access_key, 'keyid': secret_key, 'profile': {}}
error_message = 'An error occurred (101) when calling the {0} operation: Test-defined error'
error_content = {
  'Error': {
    'Code': 101,
    'Message': "Test-defined error"
  }
}

api_ret = dict(description=u'{\n    "context": "See deployment or stage description",\n    "provisioned_by": "Salt boto_apigateway.present State"\n}',
               createdDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
               id=u'vni0vq8wzi',
               name=u'unit test api')

no_apis_ret = {u'items': []}

apis_ret = {u'items': [api_ret]}

mock_model_ret = dict(contentType=u'application/json',
                      description=u'mock model',
                      id=u'123abc',
                      name=u'mock model',
                      schema=(u'{\n'
                              u'    "$schema": "http://json-schema.org/draft-04/schema#",\n'
                              u'    "properties": {\n'
                              u'        "field": {\n'
                              u'            "type": "string"\n'
                              u'        }\n'
                              u'    }\n'
                              u'}'))

models_ret = {u'items': [dict(contentType=u'application/json',
                              description=u'Error',
                              id=u'50nw8r',
                              name=u'Error',
                              schema=(u'{\n'
                                      u'    "$schema": "http://json-schema.org/draft-04/schema#",\n'
                                      u'    "properties": {\n'
                                      u'        "code": {\n'
                                      u'            "format": "int32",\n'
                                      u'            "type": "integer"\n'
                                      u'        },\n'
                                      u'        "fields": {\n'
                                      u'            "type": "string"\n'
                                      u'        },\n'
                                      u'        "message": {\n'
                                      u'            "type": "string"\n'
                                      u'        }\n'
                                      u'    },\n'
                                      u'    "title": "Error Schema",\n'
                                      u'    "type": "object"\n'
                                      u'}')),
                         dict(contentType=u'application/json',
                              description=u'User',
                              id=u'terlnw',
                              name=u'User',
                              schema=(u'{\n'
                                      u'    "$schema": "http://json-schema.org/draft-04/schema#",\n'
                                      u'    "properties": {\n'
                                      u'        "password": {\n'
                                      u'            "description": "A password for the new user",\n'
                                      u'            "type": "string"\n'
                                      u'        },\n'
                                      u'        "username": {\n'
                                      u'            "description": "A unique username for the user",\n'
                                      u'            "type": "string"\n'
                                      u'        }\n'
                                      u'    },\n'
                                      u'    "title": "User Schema",\n'
                                      u'    "type": "object"\n'
                                      u'}'))]}

root_resources_ret = {u'items': [dict(id=u'bgk0rk8rqb',
                                      path=u'/')]}

resources_ret = {u'items': [dict(id=u'bgk0rk8rqb',
                                 path=u'/'),
                            dict(id=u'9waiaz',
                                 parentId=u'bgk0rk8rqb',
                                 path=u'/users',
                                 pathPart=u'users',
                                 resourceMethods={u'POST': {}})]}

no_resources_ret = {u'items': []}

stage1_deployment1_ret = dict(cacheClusterEnabled=False,
                              cacheClusterSize=0.5,
                              cacheClusterStatus='NOT_AVAILABLE',
                              createdDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
                              deploymentId=u'kobnrb',
                              description=(u'{\n'
                                           u'    "current_deployment_label": {\n'
                                           u'        "api_name": "unit test api",\n'
                                           u'        "swagger_file": "temp-swagger-sample.yaml",\n'
                                           u'        "swagger_file_md5sum": "4fb17e43bab3a96e7f2410a1597cd0a5",\n'
                                           u'        "swagger_info_object": {\n'
                                           u'            "description": "salt boto apigateway unit test service",\n'
                                           u'            "title": "salt boto apigateway unit test service",\n'
                                           u'            "version": "0.0.0"\n'
                                           u'        }\n'
                                           u'    }\n'
                                           u'}'),
                              lastUpdatedDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
                              methodSettings=dict(),
                              stageName='test',
                              variables=dict())

stage1_deployment1_vars_ret = dict(cacheClusterEnabled=False,
                                   cacheClusterSize=0.5,
                                   cacheClusterStatus='NOT_AVAILABLE',
                                   createdDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
                                   deploymentId=u'kobnrb',
                                   description=(u'{\n'
                                                u'    "current_deployment_label": {\n'
                                                u'        "api_name": "unit test api",\n'
                                                u'        "swagger_file": "temp-swagger-sample.yaml",\n'
                                                u'        "swagger_file_md5sum": "4fb17e43bab3a96e7f2410a1597cd0a5",\n'
                                                u'        "swagger_info_object": {\n'
                                                u'            "description": "salt boto apigateway unit test service",\n'
                                                u'            "title": "salt boto apigateway unit test service",\n'
                                                u'            "version": "0.0.0"\n'
                                                u'        }\n'
                                                u'    }\n'
                                                u'}'),
                                   lastUpdatedDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
                                   methodSettings=dict(),
                                   stageName='test',
                                   variables={'var1': 'val1'})

stage1_deployment2_ret = dict(cacheClusterEnabled=False,
                              cacheClusterSize=0.5,
                              cacheClusterStatus='NOT_AVAILABLE',
                              createdDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
                              deploymentId=u'kobnrc',
                              description=(u'{\n'
                                           u'    "current_deployment_label": {\n'
                                           u'        "api_name": "unit test api",\n'
                                           u'        "swagger_file": "temp-swagger-sample.yaml",\n'
                                           u'        "swagger_file_md5sum": "5fd538c4336ed5c54b4bf39ddf97c661",\n'
                                           u'        "swagger_info_object": {\n'
                                           u'            "description": "salt boto apigateway unit test service",\n'
                                           u'            "title": "salt boto apigateway unit test service",\n'
                                           u'            "version": "0.0.2"\n'
                                           u'        }\n'
                                           u'    }\n'
                                           u'}'),
                              lastUpdatedDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
                              methodSettings=dict(),
                              stageName='test',
                              variables=dict())

stage2_ret = dict(cacheClusterEnabled=False,
                  cacheClusterSize=0.5,
                  cacheClusterStatus='NOT_AVAILABLE',
                  createdDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
                  deploymentId=u'kobnrb',
                  description=(u'{\n'
                               u'    "current_deployment_label": {\n'
                               u'        "api_name": "unit test api",\n'
                               u'        "swagger_file": "temp-swagger-sample.yaml",\n'
                               u'        "swagger_file_md5sum": "4fb17e43bab3a96e7f2410a1597cd0a5",\n'
                               u'        "swagger_info_object": {\n'
                               u'            "description": "salt boto apigateway unit test service",\n'
                               u'            "title": "salt boto apigateway unit test service",\n'
                               u'            "version": "0.0.0"\n'
                               u'        }\n'
                               u'    }\n'
                               u'}'),
                  lastUpdatedDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
                  methodSettings=dict(),
                  stageName='dev',
                  variables=dict())

stages_stage2_ret = {u'item': [stage2_ret]}

no_stages_ret = {u'item': []}

deployment1_ret = dict(createdDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
                       description=(u'{\n'
                                    u'    "api_name": "unit test api",\n'
                                    u'    "swagger_file": "temp-swagger-sample.yaml",\n'
                                    u'    "swagger_file_md5sum": "4fb17e43bab3a96e7f2410a1597cd0a5",\n'
                                    u'    "swagger_info_object": {\n'
                                    u'        "description": "salt boto apigateway unit test service",\n'
                                    u'        "title": "salt boto apigateway unit test service",\n'
                                    u'        "version": "0.0.0"\n'
                                    u'    }\n'
                                    u'}'),
                       id=u'kobnrb')

deployment2_ret = dict(createdDate=datetime.datetime(2015, 11, 17, 16, 33, 50),
                       description=(u'{\n'
                                    u'    "api_name": "unit test api",\n'
                                    u'    "swagger_file": "temp-swagger-sample.yaml",\n'
                                    u'    "swagger_file_md5sum": "5fd538c4336ed5c54b4bf39ddf97c661",\n'
                                    u'    "swagger_info_object": {\n'
                                    u'        "description": "salt boto apigateway unit test service",\n'
                                    u'        "title": "salt boto apigateway unit test service",\n'
                                    u'        "version": "0.0.2"\n'
                                    u'    }\n'
                                    u'}'),
                       id=u'kobnrc')

deployments_ret = {u'items': [deployment1_ret, deployment2_ret]}

function_ret = dict(FunctionName='unit_test_api_users_post',
                    Runtime='python2.7',
                    Role=None,
                    Handler='handler',
                    Description='abcdefg',
                    Timeout=5,
                    MemorySize=128,
                    CodeSha256='abcdef',
                    CodeSize=199,
                    FunctionArn='arn:lambda:us-east-1:1234:Something',
                    LastModified='yes')

method_integration_response_200_ret = dict(responseParameters={'method.response.header.Access-Control-Allow-Origin': '*'},
                                           responseTemplates={},
                                           selectionPattern='.*',
                                           statusCode='200')

method_integration_ret = dict(cacheKeyParameters={},
                              cacheNamespace='9waiaz',
                              credentials='arn:aws:iam::1234:role/apigatewayrole',
                              httpMethod='POST',
                              integrationResponses={'200': method_integration_response_200_ret},
                              requestParameters={},
                              requestTemplates={'application/json': ('#set($inputRoot = $input.path(\'$\'))'
                                                                     '{'
                                                                     '"header-params" : {#set ($map = $input.params().header)#foreach( $param in $map.entrySet() )"$param.key" : "$param.value" #if( $foreach.hasNext ), #end#end},'
                                                                     '"query-params" : {#set ($map = $input.params().querystring)#foreach( $param in $map.entrySet() )"$param.key" : "$param.value" #if( $foreach.hasNext ), #end#end},'
                                                                     '"path-params" : {#set ($map = $input.params().path)#foreach( $param in $map.entrySet() )"$param.key" : "$param.value" #if( $foreach.hasNext ), #end#end},'
                                                                     '"body-params" : $input.json(\'$\')'
                                                                     '}')},
                              type='AWS',
                              uri=('arn:aws:apigateway:us-west-2:'
                                   'lambda:path/2015-03-31/functions/arn:aws:lambda:us-west-2:1234567:'
                                   'function:unit_test_api_api_users_post/invocations'))

method_response_200_ret = dict(responseModels={'application/json': 'User'},
                               responseParameters={'method.response.header.Access-Control-Allow-Origin': False},
                               statusCode='200')

method_ret = dict(apiKeyRequired=False,
                  authorizationType='None',
                  httpMethod='POST',
                  methodIntegration=method_integration_ret,
                  methodResponses={'200': method_response_200_ret},
                  requestModels={'application/json': 'User'},
                  requestParameters={})

log = logging.getLogger(__name__)

opts = salt.config.DEFAULT_MINION_OPTS
context = {}
utils = salt.loader.utils(opts, whitelist=['boto3'], context=context)
serializers = salt.loader.serializers(opts)
funcs = salt.loader.minion_mods(opts, context=context, utils=utils, whitelist=['boto_apigateway'])
salt_states = salt.loader.states(opts=opts, functions=funcs, utils=utils, whitelist=['boto_apigateway'], serializers=serializers)


def _has_required_boto():
    '''
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    '''
    if not HAS_BOTO:
        return False
    elif LooseVersion(boto3.__version__) < LooseVersion(required_boto3_version):
        return False
    else:
        return True


class TempSwaggerFile(object):
    _tmp_swagger_dict = {'info': {'version': '0.0.0',
                                  'description': 'salt boto apigateway unit test service',
                                  'title': 'salt boto apigateway unit test service'},
                         'paths': {'/users': {'post': {'responses': {
                                                            '200': {'headers': {'Access-Control-Allow-Origin': {'type': 'string'}},
                                                                    'description': 'The username of the new user',
                                                                    'schema': {'$ref': '#/definitions/User'}}
                                                                    },
                                                       'parameters': [{'in': 'body',
                                                                       'description': 'New user details.',
                                                                       'name': 'NewUser',
                                                                       'schema': {'$ref': '#/definitions/User'}}],
                                                       'produces': ['application/json'],
                                                       'description': 'Creates a new user.\n',
                                                       'tags': ['Auth'],
                                                       'consumes': ['application/json'],
                                                       'summary': 'Registers a new user'}}},
                         'schemes': ['https'],
                         'produces': ['application/json'],
                         'basePath': '/api',
                         'host': 'rm06h9oac4.execute-api.us-west-2.amazonaws.com',
                         'definitions': {'User': {'properties': {
                                                     'username': {'type': 'string',
                                                                  'description': 'A unique username for the user'},
                                                     'password': {'type': 'string',
                                                                  'description': 'A password for the new user'}
                                                                }},
                                         'Error': {'properties': {
                                                     'fields': {'type': 'string'},
                                                     'message': {'type': 'string'},
                                                     'code': {'type': 'integer',
                                                              'format': 'int32'}
                                                                 }}},
                         'swagger': '2.0'}

    def __enter__(self):
        self.swaggerfile = 'temp-swagger-sample.yaml'
        with salt.utils.fopen(self.swaggerfile, 'w') as f:
            f.write(yaml.dump(self.swaggerdict))
        return self.swaggerfile

    def __exit__(self, objtype, value, traceback):
        os.remove(self.swaggerfile)

    def __init__(self, create_invalid_file=False):
        if create_invalid_file:
            self.swaggerdict = TempSwaggerFile._tmp_swagger_dict.copy()
            # add an invalid top level key
            self.swaggerdict['invalid_key'] = 'invalid'
            # remove one of the required keys 'schemes'
            self.swaggerdict.pop('schemes', None)
            # set swagger version to an unsupported verison 3.0
            self.swaggerdict['swagger'] = '3.0'
            # missing info object
            self.swaggerdict.pop('info', None)
        else:
            self.swaggerdict = TempSwaggerFile._tmp_swagger_dict


class BotoApiGatewayStateTestCaseBase(TestCase):
    conn = None

    # Set up MagicMock to replace the boto3 session
    def setUp(self):
        context.clear()
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters['key'] = ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(50))

        self.patcher = patch('boto3.session.Session')
        self.addCleanup(self.patcher.stop)
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn = MagicMock()
        session_instance.client.return_value = self.conn


@skipIf(HAS_BOTO is False, 'The boto module must be installed.')
@skipIf(_has_required_boto() is False, 'The boto3 module must be greater than'
                                       ' or equal to version {0}'
        .format(required_boto3_version))
@skipIf(NO_MOCK, NO_MOCK_REASON)
class BotoApiGatewayTestCase(BotoApiGatewayStateTestCaseBase, BotoApiGatewayTestCaseMixin):
    '''
    TestCase for salt.modules.boto_apigateway state.module
    '''

    def test_present_when_swagger_file_is_invalid(self):
        '''
        Tests present when the swagger file is invalid.
        '''
        result = {}
        with TempSwaggerFile(create_invalid_file=True) as swagger_file:
            result = salt_states['boto_apigateway.present'](
                        'api present',
                        'unit test api',
                        swagger_file,
                        'test',
                        False,
                        'arn:aws:iam::1234:role/apigatewayrole',
                        **conn_parameters)

        self.assertFalse(result.get('result', True))

    def test_present_when_stage_is_already_at_desired_deployment(self):
        '''
        Tests scenario where no action will be taken since we're already
        at desired state
        '''
        self.conn.get_rest_apis.return_value = apis_ret
        self.conn.get_deployment.return_value = deployment1_ret
        self.conn.get_stage.return_value = stage1_deployment1_ret
        self.conn.update_stage.side_effect = ClientError(error_content, 'update_stage should not be called')
        result = {}
        with TempSwaggerFile() as swagger_file:
            result = salt_states['boto_apigateway.present'](
                        'api present',
                        'unit test api',
                        swagger_file,
                        'test',
                        False,
                        'arn:aws:iam::1234:role/apigatewayrole',
                        **conn_parameters)
        self.assertFalse(result.get('abort'))
        self.assertTrue(result.get('current'))
        self.assertIs(result.get('result'), True)
        self.assertNotIn('update_stage should not be called', result.get('comment', ''))

    def test_present_when_stage_is_already_at_desired_deployment_and_needs_stage_variables_update(self):
        '''
        Tests scenario where the deployment is current except for the need to update stage variables
        from {} to {'var1':'val1'}
        '''
        self.conn.get_rest_apis.return_value = apis_ret
        self.conn.get_deployment.return_value = deployment1_ret
        self.conn.get_stage.return_value = stage1_deployment1_ret
        self.conn.update_stage.return_value = stage1_deployment1_vars_ret
        result = {}
        with TempSwaggerFile() as swagger_file:
            result = salt_states['boto_apigateway.present'](
                        'api present',
                        'unit test api',
                        swagger_file,
                        'test',
                        False,
                        'arn:aws:iam::1234:role/apigatewayrole',
                        stage_variables={'var1': 'val1'},
                        **conn_parameters)

        self.assertFalse(result.get('abort'))
        self.assertTrue(result.get('current'))
        self.assertIs(result.get('result'), True)

    def test_present_when_stage_exists_and_is_to_associate_to_existing_deployment(self):
        '''
        Tests scenario where we merely reassociate a stage to a pre-existing
        deployments
        '''
        self.conn.get_rest_apis.return_value = apis_ret
        self.conn.get_deployment.return_value = deployment2_ret
        self.conn.get_deployments.return_value = deployments_ret
        self.conn.get_stage.return_value = stage1_deployment2_ret
        self.conn.update_stage.return_value = stage1_deployment1_ret

        # should never get to the following calls
        self.conn.create_stage.side_effect = ClientError(error_content, 'create_stage')
        self.conn.create_deployment.side_effect = ClientError(error_content, 'create_deployment')

        result = {}
        with TempSwaggerFile() as swagger_file:
            result = salt_states['boto_apigateway.present'](
                        'api present',
                        'unit test api',
                        swagger_file,
                        'test',
                        False,
                        'arn:aws:iam::1234:role/apigatewayrole',
                        **conn_parameters)

        self.assertTrue(result.get('publish'))
        self.assertIs(result.get('result'), True)
        self.assertFalse(result.get('abort'))
        self.assertTrue(result.get('changes', {}).get('new', [{}])[0])

    def test_present_when_stage_is_to_associate_to_new_deployment(self):
        '''
        Tests creation of a new api/model/resource given nothing has been created previously
        '''
        # no api existed
        self.conn.get_rest_apis.return_value = no_apis_ret
        # create top level api
        self.conn.create_rest_api.return_value = api_ret
        # no models existed in the newly created api
        self.conn.get_model.side_effect = ClientError(error_content, 'get_model')
        # create model return values
        self.conn.create_model.return_value = mock_model_ret
        # various paths/resources already created
        self.conn.get_resources.return_value = resources_ret
        # the following should never be called
        self.conn.create_resource.side_effect = ClientError(error_content, 'create_resource')

        # create api method for POST
        self.conn.put_method.return_value = method_ret
        # create api method integration for POST
        self.conn.put_integration.return_value = method_integration_ret
        # create api method response for POST/200
        self.conn.put_method_response.return_value = method_response_200_ret
        # create api method integration response for POST
        self.conn.put_intgration_response.return_value = method_integration_response_200_ret

        result = {}
        with patch.dict(funcs, {'boto_lambda.describe_function': MagicMock(return_value={'function': function_ret})}):
            with TempSwaggerFile() as swagger_file:
                result = salt_states['boto_apigateway.present'](
                            'api present',
                            'unit test api',
                            swagger_file,
                            'test',
                            False,
                            'arn:aws:iam::1234:role/apigatewayrole',
                            **conn_parameters)

        self.assertIs(result.get('result'), True)
        self.assertIs(result.get('abort'), None)

    def test_present_when_stage_associating_to_new_deployment_errored_on_api_creation(self):
        '''
        Tests creation of a new api/model/resource given nothing has been created previously,
        and we failed on creating the top level api object.
        '''
        # no api existed
        self.conn.get_rest_apis.return_value = no_apis_ret
        # create top level api
        self.conn.create_rest_api.side_effect = ClientError(error_content, 'create_rest_api')

        result = {}
        with TempSwaggerFile() as swagger_file:
            result = salt_states['boto_apigateway.present'](
                        'api present',
                        'unit test api',
                        swagger_file,
                        'test',
                        False,
                        'arn:aws:iam::1234:role/apigatewayrole',
                        **conn_parameters)

        self.assertIs(result.get('abort'), True)
        self.assertIs(result.get('result'), False)
        self.assertIn('create_rest_api', result.get('comment', ''))

    def test_present_when_stage_associating_to_new_deployment_errored_on_model_creation(self):
        '''
        Tests creation of a new api/model/resource given nothing has been created previously,
        and we failed on creating the models after successful creation of top level api object.
        '''
        # no api existed
        self.conn.get_rest_apis.return_value = no_apis_ret
        # create top level api
        self.conn.create_rest_api.return_value = api_ret
        # no models existed in the newly created api
        self.conn.get_model.side_effect = ClientError(error_content, 'get_model')
        # create model return error
        self.conn.create_model.side_effect = ClientError(error_content, 'create_model')

        result = {}
        with TempSwaggerFile() as swagger_file:
            result = salt_states['boto_apigateway.present'](
                        'api present',
                        'unit test api',
                        swagger_file,
                        'test',
                        False,
                        'arn:aws:iam::1234:role/apigatewayrole',
                        **conn_parameters)

        self.assertIs(result.get('abort'), True)
        self.assertIs(result.get('result'), False)
        self.assertIn('create_model', result.get('comment', ''))

    def test_present_when_stage_associating_to_new_deployment_errored_on_resource_creation(self):
        '''
        Tests creation of a new api/model/resource given nothing has been created previously,
        and we failed on creating the resource (paths) after successful creation of top level api/model
        objects.
        '''
        # no api existed
        self.conn.get_rest_apis.return_value = no_apis_ret
        # create top level api
        self.conn.create_rest_api.return_value = api_ret
        # no models existed in the newly created api
        self.conn.get_model.side_effect = ClientError(error_content, 'get_model')
        # create model return values
        self.conn.create_model.return_value = mock_model_ret
        # get resources has nothing intiially except to the root path '/'
        self.conn.get_resources.return_value = root_resources_ret
        # create resources return error
        self.conn.create_resource.side_effect = ClientError(error_content, 'create_resource')
        result = {}
        with TempSwaggerFile() as swagger_file:
            result = salt_states['boto_apigateway.present'](
                        'api present',
                        'unit test api',
                        swagger_file,
                        'test',
                        False,
                        'arn:aws:iam::1234:role/apigatewayrole',
                        **conn_parameters)
        self.assertIs(result.get('abort'), True)
        self.assertIs(result.get('result'), False)
        self.assertIn('create_resource', result.get('comment', ''))

    def test_present_when_stage_associating_to_new_deployment_errored_on_put_method(self):
        '''
        Tests creation of a new api/model/resource given nothing has been created previously,
        and we failed on adding a post method to the resource after successful creation of top level
        api, model, resource objects.
        '''
        # no api existed
        self.conn.get_rest_apis.return_value = no_apis_ret
        # create top level api
        self.conn.create_rest_api.return_value = api_ret
        # no models existed in the newly created api
        self.conn.get_model.side_effect = ClientError(error_content, 'get_model')
        # create model return values
        self.conn.create_model.return_value = mock_model_ret
        # various paths/resources already created
        self.conn.get_resources.return_value = resources_ret
        # the following should never be called
        self.conn.create_resource.side_effect = ClientError(error_content, 'create_resource')

        # create api method for POST
        self.conn.put_method.side_effect = ClientError(error_content, 'put_method')

        result = {}
        with patch.dict(funcs, {'boto_lambda.describe_function': MagicMock(return_value={'function': function_ret})}):
            with TempSwaggerFile() as swagger_file:
                result = salt_states['boto_apigateway.present'](
                            'api present',
                            'unit test api',
                            swagger_file,
                            'test',
                            False,
                            'arn:aws:iam::1234:role/apigatewayrole',
                            **conn_parameters)

        self.assertIs(result.get('abort'), True)
        self.assertIs(result.get('result'), False)
        self.assertIn('put_method', result.get('comment', ''))

    def test_present_when_stage_associating_to_new_deployment_errored_on_lambda_function_lookup(self):
        '''
        Tests creation of a new api/model/resource given nothing has been created previously,
        and we failed on adding a post method due to a lamda look up failure after successful
        creation of top level api, model, resource objects.
        '''
        # no api existed
        self.conn.get_rest_apis.return_value = no_apis_ret
        # create top level api
        self.conn.create_rest_api.return_value = api_ret
        # no models existed in the newly created api
        self.conn.get_model.side_effect = ClientError(error_content, 'get_model')
        # create model return values
        self.conn.create_model.return_value = mock_model_ret
        # various paths/resources already created
        self.conn.get_resources.return_value = resources_ret
        # the following should never be called
        self.conn.create_resource.side_effect = ClientError(error_content, 'create_resource')
        # create api method for POST
        self.conn.put_method.return_value = method_ret
        # create api method integration for POST
        self.conn.put_integration.side_effect = ClientError(error_content, 'put_integration should not be invoked')

        result = {}
        with patch.dict(funcs, {'boto_lambda.describe_function': MagicMock(return_value={'error': 'no such lambda'})}):
            with TempSwaggerFile() as swagger_file:
                result = salt_states['boto_apigateway.present'](
                            'api present',
                            'unit test api',
                            swagger_file,
                            'test',
                            False,
                            'arn:aws:iam::1234:role/apigatewayrole',
                            **conn_parameters)

        self.assertIs(result.get('result'), False)
        self.assertNotIn('put_integration should not be invoked', result.get('comment', ''))
        self.assertIn('not find lambda function', result.get('comment', ''))

    def test_present_when_stage_associating_to_new_deployment_errored_on_put_integration(self):
        '''
        Tests creation of a new api/model/resource given nothing has been created previously,
        and we failed on adding an integration for the post method to the resource after
        successful creation of top level api, model, resource objects.
        '''
        # no api existed
        self.conn.get_rest_apis.return_value = no_apis_ret
        # create top level api
        self.conn.create_rest_api.return_value = api_ret
        # no models existed in the newly created api
        self.conn.get_model.side_effect = ClientError(error_content, 'get_model')
        # create model return values
        self.conn.create_model.return_value = mock_model_ret
        # various paths/resources already created
        self.conn.get_resources.return_value = resources_ret
        # the following should never be called
        self.conn.create_resource.side_effect = ClientError(error_content, 'create_resource')

        # create api method for POST
        self.conn.put_method.return_value = method_ret
        # create api method integration for POST
        self.conn.put_integration.side_effect = ClientError(error_content, 'put_integration')

        result = {}
        with patch.dict(funcs, {'boto_lambda.describe_function': MagicMock(return_value={'function': function_ret})}):
            with TempSwaggerFile() as swagger_file:
                result = salt_states['boto_apigateway.present'](
                            'api present',
                            'unit test api',
                            swagger_file,
                            'test',
                            False,
                            'arn:aws:iam::1234:role/apigatewayrole',
                            **conn_parameters)

        self.assertIs(result.get('abort'), True)
        self.assertIs(result.get('result'), False)
        self.assertIn('put_integration', result.get('comment', ''))

    def test_present_when_stage_associating_to_new_deployment_errored_on_put_method_response(self):
        '''
        Tests creation of a new api/model/resource given nothing has been created previously,
        and we failed on adding a method response for the post method to the resource after
        successful creation of top level api, model, resource objects.
        '''
        # no api existed
        self.conn.get_rest_apis.return_value = no_apis_ret
        # create top level api
        self.conn.create_rest_api.return_value = api_ret
        # no models existed in the newly created api
        self.conn.get_model.side_effect = ClientError(error_content, 'get_model')
        # create model return values
        self.conn.create_model.return_value = mock_model_ret
        # various paths/resources already created
        self.conn.get_resources.return_value = resources_ret
        # the following should never be called
        self.conn.create_resource.side_effect = ClientError(error_content, 'create_resource')

        # create api method for POST
        self.conn.put_method.return_value = method_ret
        # create api method integration for POST
        self.conn.put_integration.return_value = method_integration_ret
        # create api method response for POST/200
        self.conn.put_method_response.side_effect = ClientError(error_content, 'put_method_response')

        result = {}
        with patch.dict(funcs, {'boto_lambda.describe_function': MagicMock(return_value={'function': function_ret})}):
            with TempSwaggerFile() as swagger_file:
                result = salt_states['boto_apigateway.present'](
                            'api present',
                            'unit test api',
                            swagger_file,
                            'test',
                            False,
                            'arn:aws:iam::1234:role/apigatewayrole',
                            **conn_parameters)

        self.assertIs(result.get('abort'), True)
        self.assertIs(result.get('result'), False)
        self.assertIn('put_method_response', result.get('comment', ''))

    def test_present_when_stage_associating_to_new_deployment_errored_on_put_integration_response(self):
        '''
        Tests creation of a new api/model/resource given nothing has been created previously,
        and we failed on adding an integration response for the post method to the resource after
        successful creation of top level api, model, resource objects.
        '''
        # no api existed
        self.conn.get_rest_apis.return_value = no_apis_ret
        # create top level api
        self.conn.create_rest_api.return_value = api_ret
        # no models existed in the newly created api
        self.conn.get_model.side_effect = ClientError(error_content, 'get_model')
        # create model return values
        self.conn.create_model.return_value = mock_model_ret
        # various paths/resources already created
        self.conn.get_resources.return_value = resources_ret
        # the following should never be called
        self.conn.create_resource.side_effect = ClientError(error_content, 'create_resource')

        # create api method for POST
        self.conn.put_method.return_value = method_ret
        # create api method integration for POST
        self.conn.put_integration.return_value = method_integration_ret
        # create api method response for POST/200
        self.conn.put_method_response.return_value = method_response_200_ret
        # create api method integration response for POST
        self.conn.put_integration_response.side_effect = ClientError(error_content, 'put_integration_response')

        result = {}
        with patch.dict(funcs, {'boto_lambda.describe_function': MagicMock(return_value={'function': function_ret})}):
            with TempSwaggerFile() as swagger_file:
                result = salt_states['boto_apigateway.present'](
                            'api present',
                            'unit test api',
                            swagger_file,
                            'test',
                            False,
                            'arn:aws:iam::1234:role/apigatewayrole',
                            **conn_parameters)

        self.assertIs(result.get('abort'), True)
        self.assertIs(result.get('result'), False)
        self.assertIn('put_integration_response', result.get('comment', ''))

    def test_absent_when_rest_api_does_not_exist(self):
        '''
        Tests scenario where the given api_name does not exist, absent state should return True
        with no changes.
        '''
        self.conn.get_rest_apis.return_value = apis_ret
        self.conn.get_stage.side_effect = ClientError(error_content, 'get_stage should not be called')

        result = salt_states['boto_apigateway.absent'](
                    'api present',
                    'no_such_rest_api',
                    'no_such_stage',
                    nuke_api=False,
                    **conn_parameters)

        self.assertIs(result.get('result'), True)
        self.assertNotIn('get_stage should not be called', result.get('comment', ''))
        self.assertEqual(result.get('changes'), {})

    def test_absent_when_stage_is_invalid(self):
        '''
        Tests scenario where the stagename doesn't exist
        '''
        self.conn.get_rest_apis.return_value = apis_ret
        self.conn.get_stage.return_value = stage1_deployment1_ret
        self.conn.delete_stage.side_effect = ClientError(error_content, 'delete_stage')

        result = salt_states['boto_apigateway.absent'](
                    'api present',
                    'unit test api',
                    'no_such_stage',
                    nuke_api=False,
                    **conn_parameters)

        self.assertTrue(result.get('abort', False))

    def test_absent_when_stage_is_valid_and_only_one_stage_is_associated_to_deployment(self):
        '''
        Tests scenario where the stagename exists
        '''
        self.conn.get_rest_apis.return_value = apis_ret
        self.conn.get_stage.return_value = stage1_deployment1_ret
        self.conn.delete_stage.return_value = {'ResponseMetadata': {'HTTPStatusCode': 200, 'RequestId': '2d31072c-9d15-11e5-9977-6d9fcfda9c0a'}}
        self.conn.get_stages.return_value = no_stages_ret
        self.conn.delete_deployment.return_value = {'ResponseMetadata': {'HTTPStatusCode': 200, 'RequestId': '2d31072c-9d15-11e5-9977-6d9fcfda9c0a'}}

        result = salt_states['boto_apigateway.absent'](
                    'api present',
                    'unit test api',
                    'test',
                    nuke_api=False,
                    **conn_parameters)

        self.assertTrue(result.get('result', False))

    def test_absent_when_stage_is_valid_and_two_stages_are_associated_to_deployment(self):
        '''
        Tests scenario where the stagename exists and there are two stages associated with same deployment
        '''
        self.conn.get_rest_apis.return_value = apis_ret
        self.conn.get_stage.return_value = stage1_deployment1_ret
        self.conn.delete_stage.return_value = {'ResponseMetadata': {'HTTPStatusCode': 200, 'RequestId': '2d31072c-9d15-11e5-9977-6d9fcfda9c0a'}}
        self.conn.get_stages.return_value = stages_stage2_ret

        result = salt_states['boto_apigateway.absent'](
                    'api present',
                    'unit test api',
                    'test',
                    nuke_api=False,
                    **conn_parameters)

        self.assertTrue(result.get('result', False))

    def test_absent_when_failing_to_delete_a_deployment_no_longer_associated_with_any_stages(self):
        '''
        Tests scenario where stagename exists and is deleted, but a failure occurs when trying to delete
        the deployment which is no longer associated to any other stages
        '''
        self.conn.get_rest_apis.return_value = apis_ret
        self.conn.get_stage.return_value = stage1_deployment1_ret
        self.conn.delete_stage.return_value = {'ResponseMetadata': {'HTTPStatusCode': 200, 'RequestId': '2d31072c-9d15-11e5-9977-6d9fcfda9c0a'}}
        self.conn.get_stages.return_value = no_stages_ret
        self.conn.delete_deployment.side_effect = ClientError(error_content, 'delete_deployment')

        result = salt_states['boto_apigateway.absent'](
                    'api present',
                    'unit test api',
                    'test',
                    nuke_api=False,
                    **conn_parameters)

        self.assertTrue(result.get('abort', False))

    def test_absent_when_nuke_api_and_no_more_stages_deployments_remain(self):
        '''
        Tests scenario where the stagename exists and there are no stages associated with same deployment,
        the api would be deleted.
        '''
        self.conn.get_rest_apis.return_value = apis_ret
        self.conn.get_stage.return_value = stage1_deployment1_ret
        self.conn.delete_stage.return_value = {'ResponseMetadata': {'HTTPStatusCode': 200, 'RequestId': '2d31072c-9d15-11e5-9977-6d9fcfda9c0a'}}
        self.conn.get_stages.return_value = no_stages_ret
        self.conn.get_deployments.return_value = deployments_ret
        self.conn.delete_rest_api.return_value = {'ResponseMetadata': {'HTTPStatusCode': 200, 'RequestId': '2d31072c-9d15-11e5-9977-6d9fcfda9c0a'}}

        result = salt_states['boto_apigateway.absent'](
                    'api present',
                    'unit test api',
                    'test',
                    nuke_api=True,
                    **conn_parameters)

        self.assertIs(result.get('result'), True)
        self.assertIsNot(result.get('abort'), True)
        self.assertIs(result.get('changes', {}).get('new', [{}])[0].get('delete_api', {}).get('deleted'), True)

    def test_absent_when_nuke_api_and_other_stages_deployments_exist(self):
        '''
        Tests scenario where the stagename exists and there are two stages associated with same deployment,
        though nuke_api is requested, due to remaining deployments, we will not call the delete_rest_api call.
        '''
        self.conn.get_rest_apis.return_value = apis_ret
        self.conn.get_stage.return_value = stage1_deployment1_ret
        self.conn.delete_stage.return_value = {'ResponseMetadata': {'HTTPStatusCode': 200, 'RequestId': '2d31072c-9d15-11e5-9977-6d9fcfda9c0a'}}
        self.conn.get_stages.return_value = stages_stage2_ret
        self.conn.get_deployments.return_value = deployments_ret
        self.conn.delete_rest_api.side_effect = ClientError(error_content, 'unexpected_api_delete')

        result = salt_states['boto_apigateway.absent'](
                    'api present',
                    'unit test api',
                    'test',
                    nuke_api=True,
                    **conn_parameters)

        self.assertIs(result.get('result'), True)
        self.assertIsNot(result.get('abort'), True)
